import yaml
from pathlib import Path

current_dir = Path(__file__).resolve().parent
def load_config():
    with open(current_dir / "config.yaml", 'r') as stream:
        return yaml.safe_load(stream)

# Load the configuration when the package is imported
config = load_config()


from abc import ABC, abstractmethod

class Plotter(ABC):
    @abstractmethod
    def create_plot(self, data, nested):
        pass

    @abstractmethod
    def add_to_subplot(self, fig, position):
        pass

    @abstractmethod
    def get_plot(self):
        pass

    @abstractmethod
    def style(self):
        pass


from tsunami_ip_utils.readers import read_integral_indices
from tsunami_ip_utils.integral_indices import calculate_E
import numpy as np
from uncertainties import unumpy
import pandas as pd
from pandas import DataFrame as df
from pathlib import Path
from tsunami_ip_utils.perturbations import generate_points
from typing import List, Dict, Tuple, Any
from tsunami_ip_utils.viz.scatter_plot import EnhancedPlotlyFigure
from tsunami_ip_utils.viz.plot_utils import generate_plot_objects_array_from_perturbations, generate_plot_objects_from_array_contributions
from tsunami_ip_utils.integral_indices import get_uncertainty_contributions, calculate_E_contributions
from tsunami_ip_utils.viz import matrix_plot
import inspect

def comparison(tsunami_ip_output_filename: Path, application_filenames: List[Path], 
               experiment_filenames: List[Path]) -> Dict[str, df]:
    """Function that compares the calculated similarity parameter E with the TSUNAMI-IP output for each application with each
    experiment. The comparison is done for the nominal values and the uncertainties of the E values. In addition, the
    difference between manually calculated uncertainties and automatically calculated uncertainties (i.e. via the uncertainties
    package) is also calculated. The results are returned as a pandas DataFrame.
    
    Parameters
    ----------
    tsunami_ip_output_filename
        Path to the TSUNAMI-IP output file.
    application_filenames
        Paths to the application sdf files.
    experiment_filenames
        Paths to the experiment sdf files.
    
    Returns
    -------
        Dictionary of pandas DataFrames for each type of E index. The DataFrames contain the calculated E values,
        the manual uncertainties, the TSUNAMI-IP values, the relative difference in the mean, and the relative difference
        in the manual uncertainty. The DataFrames are indexed by the experiment number and the columns are a MultiIndex
        with the application number as the main index and the attributes as the subindex."""
    
    # First perform the manual calculations for each type of E index
    E_types = ['total', 'fission', 'capture', 'scatter']
    E = {}
    for E_type in E_types + ['total_manual', 'fission_manual', 'capture_manual', 'scatter_manual']:
        if 'manual' in E_type: # Manual uncertainty propagation
            if E_type.replace('_manual', '') == 'total':
                E[E_type] = calculate_E(application_filenames, experiment_filenames, reaction_type='all', uncertainties='manual')
            elif E_type.replace('_manual', '') == 'scatter':
                E[E_type] = calculate_E(application_filenames, experiment_filenames, reaction_type='elastic', uncertainties='manual')
            else:
                E[E_type] = calculate_E(application_filenames, experiment_filenames, reaction_type=E_type.replace('_manual', ''), uncertainties='manual')
        else: # Automatic uncertainty propagation
            if E_type == 'total':
                E[E_type] = calculate_E(application_filenames, experiment_filenames, reaction_type='all')
            elif E_type == 'scatter':
                E[E_type] = calculate_E(application_filenames, experiment_filenames, reaction_type='elastic')
            else:
                E[E_type] = calculate_E(application_filenames, experiment_filenames, reaction_type=E_type)

    print("Done with calculations")

    # Now read the tsunami_ip output
    tsunami_ip_output = read_integral_indices(tsunami_ip_output_filename)

    # Compare the nominal values
    E_diff = {}
    for E_type in E_types:
        E_diff[E_type] = np.abs(unumpy.nominal_values(E[E_type]) - unumpy.nominal_values(tsunami_ip_output[f"E_{E_type}"])) \
                            / unumpy.nominal_values(tsunami_ip_output[f"E_{E_type}"])

    # Compare the calculated (manual) uncertainty with the TSUNAMI-IP uncertainty
    E_diff_unc = {}
    for E_type in E_types:
        E_diff_unc[E_type] = np.abs( unumpy.std_devs(E[E_type + '_manual']) - unumpy.std_devs(tsunami_ip_output[f"E_{E_type}"]) ) \
                                / unumpy.std_devs(tsunami_ip_output[f"E_{E_type}"])

    # -----------------------------------------
    # Format the results as a pandas DataFrame
    # -----------------------------------------

    # Create a MultiIndex for columns
    num_experiments, num_applications = np.shape(E['total'])

    columns = pd.MultiIndex.from_product([
        np.arange(1, num_applications + 1),  # Main column indices
        ['Calculated', 'Manual Uncertainty', 'TSUNAMI-IP', 'Relative Difference in Mean', 'Relative Difference in Manual Uncertainty']  # Subcolumns
    ], names=['Application Number', 'Attribute'])

    # Initialize DataFrame
    data = {}

    print("Creating pandas dataframes")

    # Create a pandas DataFrame for each type of E index
    for E_type in E_types:
        data[E_type] = pd.DataFrame(index=pd.Index(np.arange(1, num_experiments + 1), name='Experiment Number'), columns=columns)

        # Populate DataFrame
        for application_index in range(num_applications):
            for experiment_index in range(num_experiments):
                # Now write the data to the DataFrame
                data[E_type].loc[experiment_index + 1, (application_index + 1, 'Calculated')] = \
                    f"{E[E_type][experiment_index, application_index].n:1.3E}+/-{E[E_type][experiment_index, application_index].s:1.2E}"
                data[E_type].loc[experiment_index + 1, (application_index + 1, 'Manual Uncertainty')] = \
                    f"{E[E_type + '_manual'][experiment_index, application_index].s:1.2E}"
                data[E_type].loc[experiment_index + 1, (application_index + 1, 'TSUNAMI-IP')] = \
                    f"{tsunami_ip_output[f'E_{E_type}'][experiment_index, application_index].n:1.3E}+/-{tsunami_ip_output[f'E_{E_type}'][experiment_index, application_index].s:1.2E}"
                data[E_type].loc[experiment_index + 1, (application_index + 1, 'Relative Difference in Mean')] = f"{E_diff[E_type][experiment_index, application_index]:1.4E}"
                data[E_type].loc[experiment_index + 1, (application_index + 1, 'Relative Difference in Manual Uncertainty')] = f"{E_diff_unc[E_type][experiment_index, application_index]:1.4E}"

    return data

def _update_annotation(fig: EnhancedPlotlyFigure, integral_index: float, index_name: str
                       ) -> Tuple[EnhancedPlotlyFigure, float, float]:
    """Update the annotation on the plot to include the TSUNAMI-IP c_k value and the percent difference from the
    Pearson correlation coefficient. If the percent difference is greater than 5%, the annotation will be colored red.
    
    Parameters
    ----------
    fig
        The plotly figure object.
    integral_index
        The TSUNAMI-IP integral_index value.
    index_name
        The name of the integral index.
        
    Returns
    -------
        * fig
            The plotly figure object with the updated annotation.
        * calculated_value
            The calculated value of the integral index.
        * percent_difference
            The percent difference between the TSUNAMI-IP integral_index value and the Pearson correlation coefficient.
    """
    summary_stats_annotation = fig.layout.annotations[0]
    calculated_value = fig.statistics['pearson']
    percent_difference = (integral_index - calculated_value)/integral_index * 100
    summary_stats_annotation.text += f"<br>TSUNAMI-IP {index_name}: <b>{integral_index}</b><br>Percent Difference: <b>{percent_difference}</b>%"
    
    if percent_difference > 5:
        summary_stats_annotation.update(bordercolor='red')
    return fig,  calculated_value, percent_difference

def correlation_comparison(integral_index_matrix: unumpy.uarray, integral_index_name: str, application_files: List[Path], 
                           experiment_files: List[Path], method: str, base_library: Path=None, perturbation_factors: Path=None, 
                           num_perturbations: int=None) -> Tuple[pd.DataFrame, Any]:
    """Function that compares the calculated similarity parameter C_k (calculated using the cross section sampling method) 
    with the TSUNAMI-IP output for each application and each experiment. NOTE: that the experiment sdfs and application sdfs
    must correspond with those in hte TSUNAMI-IP input file.

    Notes
    -----
    If the chosen method is uncertainty_contributions_nuclide or uncertainty_contributions_nuclide_reaction, the required input
    files are TSUNAMI ``.out`` files. Otherwise, the required input files are TSUNAMI ``.sdf`` files.
    
    Parameters
    ----------
    integral_index_matrix:
        The matrix representing the given integral index. Expected shape is ``(num_applications, num_experiments)``.
    integral_index_name:
        The name of the integral index (used for selecting the method for the plot). Allowed values of ``'c_k'``, ``'E'``.
    application_files
        Paths to the input files for the application (required by the chosen method, either TSUNAMI ``.out`` files or 
        TSUNAMI ``.sdf`` files).
    experiment_files
        Paths to the input files for the experiment (required by the chosen method, either TSUNAMI ``.out`` files or 
        TSUNAMI ``.sdf`` files).
    method
        The method for visualizing the given integral index. Allowed values of ``'perturbation'``, 
        ``'uncertainty_contributions_nuclide'``, ``'uncertainty_contributions_nuclide_reaction'``, 
        ``'E_contributions_nuclide'``, ``'E_contributions_nuclide_reaction'``, ``'c_k_contributions'``. 
    base_library
        Path to the base library
    perturbation_factors
        Path to the perturbation factors.
    num_perturbations
        Number of perturbations to generate.
    
    Returns
    -------
        * comparisons
            A pandas DataFrame containing the calculated integral index values, the TSUNAMI-IP values, and the percent
            difference between the two values.
        * matrix_plot
            The matrix plot object containing the integral index values and the percent difference.
    """

    # ===================================
    # Perform checks on input parameters
    # ===================================

    # Check for consistent dimensions of the integral index matrix
    num_applications, num_experiments = np.shape(integral_index_matrix)
    if num_applications != len(application_files) or num_experiments != len(experiment_files):
        raise ValueError("The dimensions of the integral index matrix do not match the number of applications and experiments.")

    # Check for missing input parameters or inconsistent method
    missing_perturbation_parameters = any([base_library is None, perturbation_factors is None,
                                             num_perturbations is None])
    if method == 'perturbation' and missing_perturbation_parameters:
        raise ValueError("The method is 'perturbation', and some of the required additional parameters are missing.")

    method_for_calculating_c_k = method in ['uncertainty_contributions_nuclide', 'uncertainty_contributions_nuclide_reaction',
                                                'perturbation', 'c_k_contributions']
    method_for_calculating_E = method in ['E_contributions_nuclide', 'E_contributions_nuclide_reaction']
    if integral_index_name == 'c_k' and not method_for_calculating_c_k:
        raise ValueError("The integral index name is 'c_k', but a method for calculating c_k was not selected, instead"
                         f"the method selected was {method}. Please select a method for calculating c_k.")
    if integral_index_name == 'E' and not method_for_calculating_E:
        raise ValueError("The integral index name is 'E', but the method selected was not 'E_contributions'. Please"
                         f"the method selected was {method}. Please select a method for calculating E.")


    # ===================================
    # Generate plots for the matrix plot
    # ===================================
    match method:
        case 'perturbation':
            points_array = generate_points(application_files, experiment_files, base_library, perturbation_factors, 
                                           num_perturbations)
            plot_objects_array = generate_plot_objects_array_from_perturbations(points_array)

        case 'uncertainty_contributions_nuclide':
            contributions_nuclide, _ = get_uncertainty_contributions(application_files, experiment_files)
            plot_objects_array = generate_plot_objects_from_array_contributions(contributions_nuclide, '%Δk/k')

        case 'uncertainty_contributions_nuclide_reaction':
            _, contributions_nuclide_reaction = get_uncertainty_contributions(application_files, experiment_files)
            plot_objects_array = generate_plot_objects_from_array_contributions(contributions_nuclide_reaction, '%Δk/k')

        case 'E_contributions_nuclide':
            contributions_nuclide, _ =  calculate_E_contributions(application_files, experiment_files)
            plot_objects_array = generate_plot_objects_from_array_contributions(contributions_nuclide, 
                                                                                integral_index_name)

        case 'E_contributions_nuclide_reaction':
            _, contributions_nuclide_reaction =  calculate_E_contributions(application_files, experiment_files)
            plot_objects_array = generate_plot_objects_from_array_contributions(contributions_nuclide_reaction, 
                                                                                integral_index_name)

        case 'c_k_contributions':
            raise NotImplementedError("The method 'c_k_contributions' is not yet implemented.")
            # plot_objects_array = generate_plot_objects_from_array_contributions(integral_index_matrix, )

    # ==============================
    # Update plots with annotations
    # ==============================
    percent_differences = np.empty_like(integral_index_matrix)
    calculated_values = np.empty_like(integral_index_matrix)
    for i, row in enumerate(plot_objects_array):
        for j, plot_object in enumerate(row):
            updated_plot_object, calculated_value, percent_difference = \
                _update_annotation(plot_object, integral_index_matrix[i, j], integral_index_name)
            calculated_values[i, j] = calculated_value
            percent_differences[i, j] = percent_difference
            plot_objects_array[i, j] = updated_plot_object

    # ===================================
    # Create dataframes with comparisons
    # ===================================
    num_applications, num_experiments = np.shape(integral_index_matrix)
    columns = pd.MultiIndex.from_product([
        np.arange(1, num_applications + 1),  # Main column indices
        ['Calculated', 'TSUNAMI-IP', 'Percent Difference']  # Subcolumns
    ], names=['Application Number', 'Attribute'])

    comparisons = pd.DataFrame(index=pd.Index(np.arange(1, num_experiments + 1), name='Experiment Number'), columns=columns)
    # Populate DataFrame
    for application_index in range(num_applications):
        for experiment_index in range(num_experiments):
            # Now write the data to the DataFrame
            comparisons.loc[experiment_index + 1, (application_index + 1, 'Calculated')] = \
                f"{calculated_values[experiment_index, application_index]:1.3E}"
            comparisons.loc[experiment_index + 1, (application_index + 1, 'TSUNAMI-IP')] = \
                f"{integral_index_matrix[experiment_index, application_index].n:1.3E}+/-{integral_index_matrix[experiment_index, application_index].s:1.2E}"
            comparisons.loc[experiment_index + 1, (application_index + 1, 'Percent Difference')] = \
                f"{percent_differences[experiment_index, application_index].n:2.2f}+/-{percent_differences[experiment_index, application_index].n:1.2E}"
    
    # ===================
    # Create matrix plot
    # ===================
    fig = matrix_plot(plot_objects_array, 'interactive')

    return comparisons, fig


from pyparsing import *
import numpy as np
from pathlib import Path
from string import Template
import tempfile
import subprocess, os
import multiprocessing
from functools import partial
import re

"""This module contains the functions necessary for processing multigroup cross sections and cross section covariance matrices."""

def parse_nuclide_reaction(filename, energy_boundaries=False):
    """Reads a multigroup cross section file produced by the extractor function and returns the energy-dependent cross sections 
    as a numpy array.
    
    Parameters
    ----------
    - filename: str The filename of the cross section file
    - energy_boundaries: bool If True, the energies at which the cross sections are defined are returned as well
    
    Returns
    -------"""
    xs = {}
    with open(filename, 'r') as f:
        data = f.read()

    # ---------------------------
    # Define grammar for parsing
    # ---------------------------

    xs_data_line = Suppress(pyparsing_common.sci_real) + pyparsing_common.sci_real + Suppress(LineEnd())

    # Note that the output is formatted such that the same cross section value is printed for both energy boundaries of the group
    # to avoid duplicating the cross section data, skip every other data line
    xs_parser = OneOrMore(xs_data_line + Suppress(xs_data_line))

    xs = np.array(xs_parser.parseString(data).asList())

    if energy_boundaries:
        # Define a parser that reads the energy boundaries of the groups
        energy_data_line = pyparsing_common.sci_real + Suppress(pyparsing_common.sci_real + LineEnd())
        energy_parser = OneOrMore(energy_data_line)
        energy_boundaries = np.unique(energy_parser.parseString(data).asList())
        return xs, energy_boundaries
    else:
        return xs

def parse_reactions_from_nuclide(filename, **kwargs):
    """Reads a set of reactions (given by the list of reaction mt's) from a dump of all reactions for a single nuclide from
    a SCALE library. Note this function requires that the dump included headers
    
    Parameters
    ----------
    - filename: str The filename of the dump file
    - reaction_mts: list of str The list of reaction MTs to read (required kwarg)
    - energy_boundaries: bool If True, the energies at which the cross sections are defined are returned as well (optional kwarg)
    
    Returns
    -------
    - dict A dictionary containing the cross sections for each reaction MT"""

    if 'reaction_mts' not in kwargs:
        raise ValueError("Missing required keyword argument: reaction_mts")
    
    reaction_mts = kwargs['reaction_mts']
    energy_boundaries = kwargs.get('energy_boundaries', False)

    if energy_boundaries:
        raise NotImplementedError("Energy boundaries are not yet supported for this function")

    with open(filename, 'r') as f:
        data = f.read()

    # ===========================
    # Define grammar for parsing
    # ===========================

    zaid = Word(nums, max=7)
    reaction_mt = Word(nums, max=4)
    fido_field = Word(nums + '$')
    fido_subfield = Word(nums + '#')

    # -------------------------------
    # Define the header line by line
    # -------------------------------
    subfield_end = Literal('t') + LineEnd()
    other_subfield_end = Literal('e') + Literal('t') + LineEnd()

    # Define a field bundle
    bundle_line1 = Suppress(fido_field) + Suppress(zaid) + Suppress(Word(nums)) + reaction_mt
    bundle_line2 = Suppress(OneOrMore(Word(nums)))
    bundle_line3 = Suppress(fido_subfield + Word(alphanums) + OneOrMore(pyparsing_common.fnumber) + other_subfield_end)
    field_bundle = bundle_line1 + bundle_line2 + bundle_line3
    
    misc_field = fido_field + Word(nums) + Word(nums)

    header = Suppress(field_bundle) + \
             field_bundle +\
             Suppress(Optional(field_bundle)) + \
             Suppress(misc_field) + \
             Suppress(fido_subfield)

    # -------------------------------------------
    # Define the cross section data line by line
    # -------------------------------------------
    xs_data_line = Suppress(pyparsing_common.sci_real) + pyparsing_common.sci_real + Suppress(LineEnd())

    # -------------------------------------------
    # Now define the total parser for a reaction
    # -------------------------------------------
    reaction_parser = header + Group(OneOrMore(xs_data_line + Suppress(xs_data_line))) + Suppress(subfield_end)

    #--------------------------------
    # Parse the data and postprocess
    #--------------------------------
    parsed_data = reaction_parser.searchString(data)
    parsed_data = { match[0]: np.array(match[1]) for match in parsed_data }
    all_mts = parsed_data.keys()
    parsed_data = { mt: data for mt, data in parsed_data.items() if mt in reaction_mts }
    if parsed_data.keys() != set(reaction_mts):
        raise ValueError(f"Not all reaction MTs were found in the data. Missing MTs: {set(reaction_mts) - set(parsed_data.keys())}. "
                         f"This nuclide has the available MTs: {list(all_mts)}")
    return parsed_data

def parse_from_total_library(filename, **kwargs):
    if 'nuclide_reaction_dict' not in kwargs:
        raise ValueError("Missing required keyword argument: nuclide_reaction_dict")
    
    if 'return_available_nuclide_reactions' in kwargs:
        return_available_nuclide_reactions = kwargs['return_available_nuclide_reactions']
    else:
        return_available_nuclide_reactions = False

    nuclide_reaction_dict = kwargs['nuclide_reaction_dict']
    energy_boundaries = kwargs.get('energy_boundaries', False)

    if energy_boundaries:
        raise NotImplementedError("Energy boundaries are not yet supported for this function")

    with open(filename, 'r') as f:
        data = f.read()

    # ===========================
    # Define regex patterns
    # ===========================

    header_pattern = re.compile(r'2\$\$\s+(\d+)\s+\d+\s+(\d+).*?(?=2##)', re.DOTALL)
    xs_data_pattern = re.compile(r'4##\s*((?:\s*[-+]?(?:\d*\.\d+|\d+\.?\d*)(?:[eE][-+]?\d+)?\s+[-+]?(?:\d*\.\d+|\d+\.?\d*)(?:[eE][-+]?\d+)?\s*\n)+)', re.MULTILINE)

    # ===========================
    # Parse the data
    # ===========================

    parsed_data_dict = {}
    all_nuclide_reactions = {}
    for match in header_pattern.finditer(data):
        nuclide = match.group(1)
        reaction = match.group(2)
        header_end = match.end()

        # Now record all available nuclide reactions
        if nuclide not in all_nuclide_reactions:
            all_nuclide_reactions[nuclide] = []
        all_nuclide_reactions[nuclide].append(reaction)

        # Check if the nuclide and reaction are in the nuclide_reaction_dict
        if nuclide in nuclide_reaction_dict and reaction in nuclide_reaction_dict[nuclide]:
            xs_data_match = xs_data_pattern.search(data, header_end)
            if xs_data_match:
                xs_data_text = xs_data_match.group(1)
                lines = xs_data_text.strip().split('\n')
                xs_data = [float(line.split()[1]) for line in lines[::2]]  # Extract second column and skip every other row
                if nuclide not in parsed_data_dict:
                    parsed_data_dict[nuclide] = {}
                parsed_data_dict[nuclide][reaction] = np.array(xs_data)

    # ========================================
    # Check for missing nuclides and reactions
    # ========================================

    nuclides_not_found = set(nuclide_reaction_dict.keys()) - set(parsed_data_dict.keys())
    reactions_not_found = {}
    for nuclide, reactions in nuclide_reaction_dict.items():
        # Skip the nuclides that aren't found
        if nuclide in nuclides_not_found:
            continue
        for reaction in reactions:
            if reaction not in parsed_data_dict[nuclide]:
                if nuclide not in reactions_not_found:
                    reactions_not_found[nuclide] = []
                reactions_not_found[nuclide].append(reaction)

    if len(nuclides_not_found) > 0 or reactions_not_found != {}:
        raise ValueError(f"Not all requested reactions were found in the data. Missing reactions: {reactions_not_found}. "
                         f"And missing nuclides: {nuclides_not_found}")


    if return_available_nuclide_reactions:
        # Remove the nuclide '0' from the list of available nuclides if it exists
        all_nuclide_reactions.pop('0', None)
        return parsed_data_dict, all_nuclide_reactions
    else:
        return parsed_data_dict

def read_nuclide_reaction_from_multigroup_library(multigroup_library_path: Path, nuclide_zaid, reaction_mt, \
                                                  parsing_function=parse_nuclide_reaction, plot_option='plot', \
                                                    energy_boundaries=False, **kwargs):
    """Uses scale to dump a binary multigroup library to a text file, and then calls the specified parsing function on the output file.
    
    Parameters
    ----------
    - multigroup_library_path: Path The path to the SCALE multigroup library file
    - nuclide_zaid: str The ZAID of the nuclide
    - reaction_mt: str The reaction MT to read
    - parsing_function: function The function to call on the output file
    - plot_option: str The plot option to use when running the MG reader
    - energy_boundaries: bool If True, the energies at which the cross sections are defined are returned as well
    
    Returns
    -------
    - an output that is the result of the parsing function"""
    # Get the directory of the current file
    current_dir = Path(__file__).parent

    # Construct the path to the input file
    file_path = current_dir / 'input_files' / 'MG_reader.inp'

    # Create a tempfile for storing the output file of the MG reader dump.
    with tempfile.NamedTemporaryFile(mode='w', delete=False) as output_file:
        output_file_path = output_file.name
    
    # Read the MG reader input template file
    with open(file_path, 'r') as f:
        template = Template(f.read())
        
    # Substitute the input file template variables
    input_file = template.safe_substitute(
        nuclide_zaid=nuclide_zaid, 
        reaction_mt=reaction_mt, 
        multigroup_library_path=multigroup_library_path,
        output_file_path=output_file_path,
        plot_option=plot_option
    )

    # Write the input file to another tempfile
    with tempfile.NamedTemporaryFile(mode='w', delete=False) as input_temp_file:
        input_temp_file.write(input_file)
        input_temp_file_path = input_temp_file.name

    # Run the executable
    command = ['scalerte', input_temp_file_path]

    proc = subprocess.Popen(command)
    proc.wait()

    # Now delete the input file
    os.remove(input_temp_file_path)

    # Read the output file
    output = parsing_function(output_file_path, energy_boundaries=energy_boundaries, **kwargs)

    # Now delete the output file
    os.remove(output_file_path)
    return output

def read_reactions_from_nuclide(multigroup_library_path: Path, nuclide_zaid, reaction_mts):
    """Function for reading a set of reactions from a given nuclide in a SCALE multigroup library.
    
    Parameters
    ----------
    - multigroup_library_path: Path The path to the SCALE multigroup library file
    - nuclide_zaid: str The ZAID of the nuclide
    - reaction_mts: list The list of reaction MTs to read"""
    output = read_nuclide_reaction_from_multigroup_library(multigroup_library_path, nuclide_zaid, reaction_mt='0', \
                                                           parsing_function=parse_reactions_from_nuclide, \
                                                            reaction_mts=reaction_mts, plot_option='fido')
    return output

def read_multigroup_xs(multigroup_library_path: Path, nuclide_zaid_reaction_dict, method='small', num_processes=None, \
                       return_available_nuclide_reactions=False):
    """Function for reading a set of reactions from a given nuclide in a SCALE multigroup library.
    
    Parameters
    ----------
    - multigroup_library_path: Path The path to the SCALE multigroup library file
    - nuclide_zaid_reaction_dict: dict A dictionary mapping nuclide ZAIDs to a list of reaction MTs to read"""

    NUCLIDE_THRESHOLD = 50 # Number of nuclides after which the large method is more performant and hence is used
    CORE_THRESHOLD = 2 # The large method is more performant if the number of cores is smaller than this number

    if num_processes is None:
        num_processes = multiprocessing.cpu_count()

    num_nuclides = len(list(nuclide_zaid_reaction_dict.keys()))
    use_small_method = ( num_nuclides < NUCLIDE_THRESHOLD ) and ( num_processes >= CORE_THRESHOLD )
    if use_small_method and not return_available_nuclide_reactions: # This method is slow but works for small amounts of nuclide reactions or a large amount of cores
        pool = multiprocessing.Pool(processes=num_processes)

        # Create a partial function with the common arguments
        read_reactions_partial = partial(read_reactions_from_nuclide, multigroup_library_path)

        # Distribute the function calls among the processes
        results = pool.starmap(read_reactions_partial, nuclide_zaid_reaction_dict.items())

        # Close the pool and wait for the processes to finish
        pool.close()
        pool.join()

        # Convert the results to a dictionary
        output = dict(zip(nuclide_zaid_reaction_dict.keys(), results))

        return output
    else: # This method is faster (as in there's less scale run overhead) but requires the entire library to be read and is serial
        # If the user wants the available nuclide reactions, then we need to create a partial function which adds the appropriate
        # keyword argument to the parsing function
        if return_available_nuclide_reactions:
            parse_function = partial(parse_from_total_library, return_available_nuclide_reactions=True)
        else:
            parse_function = parse_from_total_library
        
        output = read_nuclide_reaction_from_multigroup_library(multigroup_library_path, nuclide_zaid='0', reaction_mt='0', \
                                                        parsing_function=parse_function, \
                                                        nuclide_reaction_dict=nuclide_zaid_reaction_dict, plot_option='fido')
        return output


"""This module is used for generating cross section perturbations and combining them with the sensitivity profiles for a given application
experiment pair to generate a similarity scatter plot"""

from tsunami_ip_utils.readers import RegionIntegratedSdfReader, read_region_integrated_h5_sdf
from tsunami_ip_utils.utils import filter_redundant_reactions
from pathlib import Path
from tsunami_ip_utils.xs import read_multigroup_xs
import pickle
import os
import tempfile
from string import Template
import subprocess
import time
from tsunami_ip_utils.utils import filter_by_nuclie_reaction_dict
import multiprocessing
from multiprocessing import Pool
from tqdm import tqdm
import numpy as np
from tqdm.contrib.concurrent import process_map
from typing import List, Tuple, Dict, Union
from . import config
from numpy.typing import ArrayLike

# Number of xs perturbation samples available in SCALE
NUM_SAMPLES = config['NUM_SAMPLES']

def _generate_and_read_perturbed_library(base_library: Path, perturbation_factors: Path, sample_number: int, 
                                         all_nuclide_reactions: dict) -> dict:
    """Generates and reads perturbed multigroup cross section libraries.
    
    Parameters
    ----------
    base_library
        Path to the base cross section library.
    perturbation_factors
        Path to the perturbation factors directory (corresponding to the base library).
    sample_number
        The sample number to use for generating the perturbed library. Must be from 1 - ``NUM_SAMPLES``, where 
        ``NUM_SAMPLES`` = :globalparam:`NUM_SAMPLES` is the number of perturbation factor samples provided in 
        the user's current version of SCALE. (``0`` :math:`\\leq` ``sample_number`` :math:`\\leq` :globalparam:`NUM_SAMPLES`)
    all_nuclide_reactions
        A dictionary containing the nuclide reactions that are read from the perturbed library.
    
    Returns
    -------
        A dictionary containing the perturbed cross section libraries for each nuclide reaction."""
    # Read the SCALE input template
    current_dir = Path(__file__).parent
    template_filename = current_dir / 'input_files' / 'generate_perturbed_library.inp'
    with open(template_filename, 'r') as f:
        template = Template(f.read())

    # Open a temporary file to store the output file
    with tempfile.NamedTemporaryFile(mode='w', delete=False) as output_file:
        output_file_path = output_file.name

    # Substitute the input file template values
    input_file = template.substitute(
        base_library=str(base_library),
        perturbation_factors=str( perturbation_factors / f'Sample{sample_number}'),
        sample_number=sample_number,
        output=output_file_path
    )

    # Open a temporary file to store the input file, then one to store the output file
    with tempfile.NamedTemporaryFile(mode='w', delete=False) as input_temp_file:
        input_temp_file.write(input_file)
        input_temp_file_path = input_temp_file.name

    # Run the executable
    command = ['scalerte', input_temp_file_path]

    proc = subprocess.Popen(command)
    proc.wait()

    # Now delete the input file
    os.remove(input_temp_file_path)

    # Read the perturbed library
    perturbed_xs = read_multigroup_xs(Path(output_file_path), all_nuclide_reactions)

    # Now delete the output file
    os.remove(output_file_path)

    return perturbed_xs

def generate_points(application_path: Union[Path, List[Path]], experiment_path: Union[Path, List[Path]], base_library: Path, 
                    perturbation_factors: Path, num_perturbations: int
                    ) -> Union[ List[ Tuple[ float, float ] ], 
                               np.ndarray[ List[ Tuple[ float, float ] ] ]]:
    """Generates points for a similarity scatter plot by combining the sensitivity profiles of the application and experiment
    with the perturbed cross section libraries.

    Notes
    -----
    * This function will cache the base cross section library and the perturbed cross section libraries in the user's home 
    directory under the ``.tsunami_ip_utils_cache`` directory. If the user wishes to reset the cache, they can do so by
    setting the ``reset_cache`` parameter to ``True`` in the :func:`cache_all_libraries` function. The user can also cache
    the base and perturbed cross section libraries manually by calling the :func:`cache_all_libraries` function.

    * This function can also generate a matrix of points for a given set of experiment and applications for making a matrix plot
    this can be done by passing a list of paths for the application and experiment sensitivity profiles.

    Parameters
    ----------
    application_path
        Path(s) to the application sensitivity profile.
    experiment_path
        Path(s) to the experiment sensitivity profile.
    base_library
        Path to the base cross section library.
    perturbation_factors
        Path to the perturbation factors directory.
    num_perturbations
        Number of perturbation points to generate.
    
    Returns
    -------
        A list of points for the similarity scatter plot."""
    
    if isinstance(application_path, list) and isinstance(experiment_path, list):
        points_array = np.empty( ( len(application_path), len(experiment_path) ), dtype=object )
        for i, application in enumerate(application_path):
            for j, experiment in enumerate(experiment_path):
                points_array[i, j] = generate_points(
                    application, 
                    experiment, 
                    base_library, 
                    perturbation_factors, 
                    num_perturbations
                )
        return points_array
    elif isinstance(application_path, list) or isinstance(experiment_path, list):
        raise ValueError("Both application and experiment paths must be lists or neither.")

    # Read the sdfs for the application and experiment
    if application_path.suffix == '.sdf':
        application = RegionIntegratedSdfReader(application_path).convert_to_dict('numbers').sdf_data
        application = { nuclide_name: { reaction_name: reaction['sensitivities'] } \
                       for nuclide_name, nuclide in application.items() \
                       for reaction_name, reaction in nuclide.items() }
    elif application_path.suffix == '.h5':
        application = read_region_integrated_h5_sdf(application_path)
    else:
        raise ValueError("The application path must be either an sdf or h5 file.")
    
    if experiment_path.suffix == '.sdf':
        experiment  = RegionIntegratedSdfReader(experiment_path).convert_to_dict('numbers').sdf_data
        experiment = { nuclide_name: { reaction_name: reaction['sensitivities'] } \
                      for nuclide_name, nuclide in experiment.items() \
                      for reaction_name, reaction in nuclide.items() }
    elif experiment_path.suffix == '.h5':
        experiment = read_region_integrated_h5_sdf(experiment_path)
    else:
        raise ValueError("The experiment path must be either an sdf or h5 file.")
    

    # Filter out redundant reactions, which will introduce bias into the similarity scatter plot
    # Absorption, or "capture" as it's referred to in SCALE and total
    # redundant_reactions = ['101', '1'] 
    # application = filter_redundant_reactions(application, redundant_reactions=redundant_reactions)
    # experiment  = filter_redundant_reactions(experiment,  redundant_reactions=redundant_reactions)

    # Create a nuclide reaction dict for the application and experiment
    application_nuclide_reactions = {nuclide: list(reactions.keys()) for nuclide, reactions in application.items()}
    experiment_nuclide_reactions  = {nuclide: list(reactions.keys()) for nuclide, reactions in experiment.items()}

    # Take the union of the nuclide reactions for the application and experiment
    all_nuclide_reactions = application_nuclide_reactions.copy()
    all_nuclide_reactions.update(experiment_nuclide_reactions)

    # Make a directory to store all cached cross section libraries if it doesn't already exist
    cache_dir = Path.home() / ".tsunami_ip_utils_cache"
    if not cache_dir.exists():
        os.mkdir(cache_dir)

    # Make a directory to store the cached perturbed multigroup libraries if it doesn't already exist
    library_name = base_library.name

    # Get the base multigroup cross sections for each nuclide reaction and the list of all available nuclide reactions for
    # caching the sampled perturbed cross sections

    if not (cache_dir / f'cached_{library_name}.pkl').exists():
        base_xs, available_nuclide_reactions = read_multigroup_xs(base_library, all_nuclide_reactions, \
                                                                return_available_nuclide_reactions=True)
        # Now read all cross sections and cache the base cross section library
        base_xs = read_multigroup_xs(base_library, available_nuclide_reactions)
        with open(cache_dir / f'cached_{library_name}.pkl', 'wb') as f:
            pickle.dump(base_xs, f)
    else:
        with open(cache_dir / f'cached_{library_name}.pkl', 'rb') as f:
            base_xs = pickle.load(f)

        # Get available nuclide reactions
        available_nuclide_reactions = { nuclide: list( reactions.keys() ) for nuclide, reactions in base_xs.items() }

        # Filter out the desired nuclide reactions
        base_xs = filter_by_nuclie_reaction_dict(base_xs, all_nuclide_reactions)

    perturbed_cache = cache_dir / f'cached_{library_name}_perturbations'
    if not perturbed_cache.exists():
        os.mkdir(perturbed_cache)

    # --------------------------------
    # Main loop for generating points
    # --------------------------------
    points = []
    for i in tqdm(range(1, num_perturbations + 1), desc="Generating perturbation points"):
        # Cache the perturbed cross section libraries if not already cached
        perturbed_xs_cache = perturbed_cache / f'perturbed_xs_{i}.pkl'
        if not perturbed_xs_cache.exists():
            perturbed_xs = _generate_and_read_perturbed_library(base_library, perturbation_factors, i, \
                                                               available_nuclide_reactions)
            with open(perturbed_xs_cache, 'wb') as f:
                pickle.dump(perturbed_xs, f)
        else:
            with open(perturbed_xs_cache, 'rb') as f:
                perturbed_xs = pickle.load(f)

        # Now filter out the desired nuclide reactions
        perturbed_xs = filter_by_nuclie_reaction_dict(perturbed_xs, all_nuclide_reactions)

        # ----------------------------------------------
        # Compute S ⋅ Δσ for application and experiment
        # ----------------------------------------------
        running_total_application = 0
        running_total_experiment = 0
        for isotope, reactions in all_nuclide_reactions.items():
            for reaction in reactions:
                # First compute the cross section delta, then multiply by the sensitivity profile
                delta_xs = perturbed_xs[isotope][reaction] - base_xs[isotope][reaction]
                if isotope in application and reaction in application[isotope]:
                    running_total_application += np.dot(application[isotope][reaction], delta_xs)
                if isotope in experiment and reaction in experiment[isotope]:
                    running_total_experiment  += np.dot(experiment[isotope][reaction], delta_xs)

        points.append((running_total_application, running_total_experiment))

    return points
        

def _cache_perturbed_library(args: Tuple[int, Path, Path, int, Dict[str, List[str]], Path]) -> float:
    """Caches a single perturbed cross section library.

    Parameters
    ----------
    args
        A tuple containing all necessary components to perform the caching of a 
        perturbed library.
        
        - i (int):
            The sample number to use for generating the perturbed library.
        - base_library (Path):
            Path to the base cross section library.
        - perturbation_factors (Path):
            Path to the cross section perturbation factors (used to generate the perturbed libraries).
        - sample_number (int):
            The sample number to use for generating the perturbed library. Must be from 1 - ``NUM_SAMPLES``, where ``NUM_SAMPLES``
            is the number of perturbation factor samples provided in the user's current version of SCALE. 
            (0 :math:`\\leq` ``sample_number`` :math:`\\leq` ``NUM_SAMPLES``)
        - available_nuclide_reactions (Dict[str, List[str]]):
            A dictionary containing the nuclide reactions that are read from the perturbed library.

    Returns
    -------
        The time taken to cache the perturbed library.
    """    
    i, base_library, perturbation_factors, sample_number, available_nuclide_reactions, perturbed_cache = args
    perturbed_xs_cache = perturbed_cache / f'perturbed_xs_{i}.pkl'
    if not perturbed_xs_cache.exists():
        start = time.time()
        perturbed_xs = _generate_and_read_perturbed_library(base_library, perturbation_factors, sample_number, available_nuclide_reactions)
        with open(perturbed_xs_cache, 'wb') as f:
            pickle.dump(perturbed_xs, f)
        
        end = time.time()
        return end - start
    else:
        return 0

def cache_all_libraries(base_library: Path, perturbation_factors: Path, reset_cache: bool=False) -> None:
    """Caches the base and perturbed cross section libraries for a given base library and perturbed library paths

    Parameters
    ----------
    base_library
        Path to the base cross section library.
    perturbation_factors
        Path to the cross section perturbation factors (used to generate the perturbed libraries).
    reset_cache
        Whether to reset the cache or not (default is ``False``).
        
    Returns
    -------
        This function does not return a value and has no return type."""
    # Read the base library, use an arbitrary nuclide reaction dict just to get the available reactions
    all_nuclide_reactions = { '92235': ['18'] } # u-235 fission

    # Make a directory to store all cached cross section libraries if it doesn't already exist
    cache_dir = Path.home() / ".tsunami_ip_utils_cache"
    if not cache_dir.exists():
        os.mkdir(cache_dir)

    if reset_cache:
        # Remove all cached cross section libraries
        for f in os.listdir(cache_dir):
            if f.endswith('_perturbations'): # A perturbed library directory
                for p in os.listdir(cache_dir / f):
                    os.remove(cache_dir / f / p)
            else:
                os.remove(cache_dir / f)

    # Cache base library if not already cached (requires reading twice)
    library_name = base_library.name
    base_library_cache = cache_dir / f'cached_{library_name}.pkl'
    print("Reading base library... ")
    if not base_library_cache.exists():
        _, available_nuclide_reactions = read_multigroup_xs(base_library, all_nuclide_reactions, \
                                                                return_available_nuclide_reactions=True)

        start = time.time()
        print("Caching base library... ", end='')
        base_xs = read_multigroup_xs(base_library, available_nuclide_reactions)
        with open( base_library_cache, 'wb') as f:
            pickle.dump(base_xs, f)
        end = time.time()
        print(f"Done in {end - start} seconds")
    else:
        with open(cache_dir / f'cached_{library_name}.pkl', 'rb') as f:
            base_xs = pickle.load(f)

        # Get available nuclide reactions
        available_nuclide_reactions = { nuclide: list( reactions.keys() ) for nuclide, reactions in base_xs.items() }
        
    
    # Make a directory to store the cached perturbed multigroup libraries if it doesn't already exist
    perturbed_cache = cache_dir / f'cached_{library_name}_perturbations'
    if not perturbed_cache.exists():
        os.mkdir(perturbed_cache)

    # ------------------------------------------
    # Main loop for caching perturbed libraries
    # ------------------------------------------
    # Get the number of available cores
    num_cores = multiprocessing.cpu_count()
    
    # Use only half of the available cores
    num_processes = num_cores // 2

    # Create a pool of worker processes
    pool = Pool(processes=num_processes)

    # Create a list of arguments for each perturbed library
    args_list = [(i, base_library, perturbation_factors, i, available_nuclide_reactions, perturbed_cache)
                 for i in range(1, NUM_SAMPLES + 1)]

    # Use the pool to cache the perturbed libraries in parallel with a progress bar
    process_map(_cache_perturbed_library, args_list, max_workers=num_cores // 2, chunksize=1, desc='Caching perturbed libraries')

    # Close the pool
    pool.close()
    pool.join()


import scipy.stats as stats
import numpy as np
import matplotlib.pyplot as plt
from .base_plotter import Plotter
import plotly.express as px
import plotly.graph_objects as go
from plotly.subplots import make_subplots
import pandas as pd
import plotly.graph_objects as go
import dash
from dash import dcc, html
from dash.dependencies import Input, Output, State
import pandas as pd
import webbrowser
import os
import signal
import pickle
import threading
from .plot_utils import find_free_port
import sys, os, signal
import threading
import webbrowser
import sys
from plotly.graph_objs import Figure


class EnhancedPlotlyFigure(Figure):
    """This class wraps a plotly express figure object (intended for a scatter plot) and adds additional metadata for the
    summary statistics and linear regression data. This class is intended to be used with the ``InteractiveScatterPlotter``
    class."""
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Directly set the attributes using object's __setattr__ to bypass Plotly's checks
        object.__setattr__(self, 'statistics', None)
        object.__setattr__(self, 'regression', None)

    def __setattr__(self, name, value):
        if name in ['statistics', 'regression']:
            # Handle custom attributes internally
            object.__setattr__(self, name, value)
        else:
            # Use the super class's __setattr__ for all other attributes
            super().__setattr__(name, value)

class ScatterPlot(Plotter):
    """This class exists to add some additional functionality for calculating regressions and summary statistics that's
    common to all types of scatter plots, interactive or otherwise"""
    def get_summary_statistics(self, x, y):
        """Calculates the Pearson correlation coefficient, Spearman rank correlation coefficient, and linear regression
        parameters for the given x and y datasets. The linear regression parameters are the slope and intercept of the
        regression line. The Pearson and Spearman coefficients are also stored in the class instance as 'pearson' and
        'spearman' respectively. The slope and intercept are stored as 'slope' and 'intercept' respectively. The linear
        regression is stored as 'regression'"""
        self.regression = stats.linregress(x, y)
        self.pearson    = stats.pearsonr(x, y).statistic
        self.spearman   = stats.spearmanr(x, y).statistic
        self.slope      = self.regression.slope
        self.intercept  = self.regression.intercept

        # If the figure has been plotted (and is an enhanced plot which supports adding this metadata), add the regression 
        # and correlation statistics to the figure
        is_metadata_plot = isinstance(getattr(self, 'fig', None), EnhancedPlotlyFigure)
        if hasattr(self, 'fig') and is_metadata_plot:
            self.fig.statistics = {
                'pearson': self.pearson,
                'spearman': self.spearman
            }
            self.fig.regression = {
                'slope': self.slope,
                'intercept': self.intercept,
            }

        # Now create teh summary statistics text for figure annotation
        pearson_text = f"Pearson: <b>{self.pearson:1.6f}</b>"
        spearman_text = f"Spearman: <b>{self.spearman:1.6f}</b>"
        self.summary_stats_text = f"{pearson_text} {spearman_text}"


class ScatterPlotter(ScatterPlot):
    def __init__(self, integral_index_name, nested, plot_redundant=False, **kwargs):
        self.nested = nested
        self.index_name = integral_index_name
        self.plot_redundant = plot_redundant

    def create_plot(self, contribution_pairs, isotopes, reactions):
        self.fig, self.axs = plt.subplots()

        # Extract the x and y values from the contribution pairs
        application_points        = [ contribution[0].n for contribution in contribution_pairs ]
        application_uncertainties = [ contribution[0].s for contribution in contribution_pairs ]
        experiment_points         = [ contribution[1].n for contribution in contribution_pairs ]
        experiment_uncertainties  = [ contribution[1].s for contribution in contribution_pairs ]

        self.fig = plt.errorbar(application_points, experiment_points, xerr=application_uncertainties, \
                               yerr=experiment_uncertainties, fmt='.', capsize=5)
        
        # Linear regression
        self.get_summary_statistics(application_points, experiment_points)

        # Plot the regression line
        x = np.linspace(min(application_points), max(application_points), 100)
        y = self.slope * x + self.intercept
        self.axs.plot(x, y, 'r', label='Linear fit')

        self.axs.text(0.05, 0.95, self.summary_stats_text, transform=self.axs.transAxes, fontsize=12,
                verticalalignment='top', bbox=dict(facecolor='white', alpha=0.5))

        self.style()

    def get_plot(self):
        return self.fig, self.axs
        
    def add_to_subplot(self, fig, position):
        return fig.add_subplot(position, sharex=self.axs, sharey=self.axs)
    
    def style(self):
        title_text = f'Contributions to {self.index_name}'
        self.axs.set_title(title_text)
        self.axs.set_ylabel(f"Experiment {self.index_name} Contribution")
        self.axs.set_xlabel(f"Application {self.index_name} Contribution")
        self.axs.grid()


def load_interactive_scatter_plot(filename):
    """Loads an interactive scatter plot from a saved state file. This function is purely for convenience and is a
    wrapper of the InteractiveScatterLegend.load_state method"""
    return InteractiveScatterLegend.load_state(filename)


class InteractiveScatterPlotter(ScatterPlot):
    def __init__(self, integral_index_name, nested, plot_redundant=False, **kwargs):
        if 'interactive_legend' in kwargs.keys():
            self.interactive_legend = kwargs['interactive_legend']
        else:
            self.interactive_legend = False
        self.nested = nested
        self.index_name = integral_index_name
        self.plot_redundant = plot_redundant

    def create_plot(self, contribution_pairs, isotopes, reactions):
        self.fig = make_subplots()

        # Extract isotope and reaction pairs from the given list of isotopes and reactions
        df = self._create_scatter_data(contribution_pairs, isotopes, reactions)

        hover_data_dict = {
            'Isotope': True  # Always include Isotope
        }

        if 'Reaction' in df.columns:
            hover_data_dict['Reaction'] = True  # Include Reaction only if it exists

        # Create scatter plot with error bars using Plotly Express
        self.fig = px.scatter(
            df, 
            x=f'Application {self.index_name} Contribution', 
            y=f'Experiment {self.index_name} Contribution',
            error_x='Application Uncertainty', 
            error_y='Experiment Uncertainty',
            color='Isotope',
            labels={
                "color": "Isotope"
            },
            title=f'Contributions to {self.index_name}',
            hover_data=hover_data_dict
        )

        # Wrap the plotly express figure in a MetadataPlotly object
        self.fig = EnhancedPlotlyFigure(self.fig.to_dict())

        self.add_regression_and_stats(df)

        # Now style the plot
        self.style()

        if self.interactive_legend:
            self.fig = InteractiveScatterLegend(self, df)

    def add_regression_and_stats(self, df):
        # Calculate the linear regression and correlation statistics
        self.get_summary_statistics(df[f'Application {self.index_name} Contribution'], \
                                    df[f'Experiment {self.index_name} Contribution'])

        # Prepare data for the regression line
        x_reg = np.linspace(df[f'Application {self.index_name} Contribution'].min(), 
                            df[f'Application {self.index_name} Contribution'].max(), 100)
        y_reg = self.slope * x_reg + self.intercept

        # Convert self.fig.data to a list for mutability
        current_traces = list(self.fig.data)

        # Remove existing regression line if it exists
        traces_to_keep = [trace for trace in current_traces if not trace.name.startswith('Regression Line')]

        # Set the modified list of traces back to the figure
        self.fig.data = tuple(traces_to_keep)

        # Remove existing annotation if it exists
        if hasattr(self.fig, 'layout') and hasattr(self.fig.layout, 'annotations'):
            self.fig.layout.annotations = [ann for ann in self.fig.layout.annotations if not ann.text.startswith('Pearson')]

        # Add new linear regression to the plot
        self.fig.add_trace(go.Scatter(x=x_reg, y=y_reg, mode='lines', 
                                    name=f'Regression Line y={self.slope:1.4E}x + {self.intercept:1.4E}'))

        # Add correlation statistics to the plot
        self.fig.add_annotation(
            x=0.05, xref="paper", 
            y=0.95, yref="paper",
            text=self.summary_stats_text,
            showarrow=False, 
            font=dict(size=12),
            align='left',
            bgcolor="white", 
            opacity=0.8
        )

    def _create_scatter_data(self, contribution_pairs, isotopes, reactions):
        data = {
            f'Application {self.index_name} Contribution': [cp[0].n for cp in contribution_pairs],
            f'Experiment {self.index_name} Contribution': [cp[1].n for cp in contribution_pairs],
            'Application Uncertainty': [cp[0].s for cp in contribution_pairs],
            'Experiment Uncertainty': [cp[1].s for cp in contribution_pairs],
            'Isotope': [],
        }

        # Add nuclides and reactions (if they exist) to the data dictionary
        if reactions == []:
            for isotope in isotopes:
                data['Isotope'].append(isotope)
        else:
            data['Reaction'] = []
            for isotope in isotopes:
                for reaction in reactions:
                    data['Isotope'].append(isotope)
                    data['Reaction'].append(reaction)

        # Now filter out (0,0) points, which don't contribute to either the application or the experiment, these are
        # usually chi, nubar, or fission reactions for nonfissile isotopes that are added for consistency with the set
        # of reactions only
        data = { key: [val for val, app, exp in zip(data[key], data[f'Application {self.index_name} Contribution'], \
                    data[f'Experiment {self.index_name} Contribution']) if app != 0 or exp != 0] for key in data }

        return pd.DataFrame(data)

    def add_to_subplot(self, fig, position):
        for trace in self.fig.data:
            fig.add_trace(trace, row=position[0], col=position[1])

        # Transfer annotations
        if hasattr(self.fig, 'layout') and hasattr(self.fig.layout, 'annotations'):
            for ann in self.fig.layout.annotations:
                # Adjust annotation references to new subplot
                new_ann = ann.update(xref=f'x{position[1]}', yref=f'y{position[1]}')
                fig.add_annotation(new_ann, row=position[0], col=position[1])
        return fig

    def get_plot(self):
        return self.fig
    
    def style(self):
        title_text = f'Contributions to {self.index_name}'
        self.fig.update_layout(title_text=title_text, title_x=0.5)  # 'title_x=0.5' centers the title


class InteractivePerturbationScatterPlotter(ScatterPlot):
    def __init__(self, **kwargs):
        pass

    def create_plot(self, points):
        self.fig = make_subplots()

        # Extract isotope and reaction pairs from the given list of isotopes and reactions
        df = pd.DataFrame({
            'Application': [point[0].n for point in points],
            'Experiment': [point[1].n for point in points],
            'Application Uncertainty': [point[0].s for point in points],
            'Experiment Uncertainty': [point[1].s for point in points]
        })

        # Create scatter plot with error bars using Plotly Express
        self.fig = px.scatter(
            df, 
            x=f'Application', 
            y=f'Experiment',
            error_x='Application Uncertainty', 
            error_y='Experiment Uncertainty',
            title=f'Correlation Plot',
        )

        # Wrap the plotly express figure in a MetadataPlotly object
        self.fig = EnhancedPlotlyFigure(self.fig.to_dict())

        self.add_regression_and_stats(df)

        # Now style the plot
        self.style()

    def add_regression_and_stats(self, df):
        # Calculate the linear regression and correlation statistics
        self.get_summary_statistics(df[f'Application'], df[f'Experiment'])

        # Prepare data for the regression line
        x_reg = np.linspace(df[f'Application'].min(), df[f'Application'].max(), 100)
        y_reg = self.slope * x_reg + self.intercept

        # Convert self.fig.data to a list for mutability
        current_traces = list(self.fig.data)

        # Remove existing regression line if it exists
        traces_to_keep = [trace for trace in current_traces if not trace.name.startswith('Regression Line')]

        # Set the modified list of traces back to the figure
        self.fig.data = tuple(traces_to_keep)

        # Add new linear regression to the plot
        self.fig.add_trace(go.Scatter(x=x_reg, y=y_reg, mode='lines', 
                                    name=f'Regression Line y={self.slope:1.4E}x + {self.intercept:1.4E}'))

        # Add correlation statistics to the plot
        self.fig.add_annotation(
            x=0.05, xref="paper", 
            y=0.95, yref="paper",
            text=self.summary_stats_text,
            showarrow=False,
            align='left',
            font=dict(size=12),
            bgcolor="white", 
            opacity=0.8
        )

    def add_to_subplot(self, fig, position):
        for trace in self.fig.data:
            fig.add_trace(trace, row=position[0], col=position[1])
        return fig

    def get_plot(self):
        return self.fig
    
    def style(self):
        pass


class InteractiveScatterLegend(InteractiveScatterPlotter):
    def __init__(self, interactive_scatter_plot, df):
        self.interactive_scatter_plot = interactive_scatter_plot
        self.fig = interactive_scatter_plot.fig
        self.index_name = interactive_scatter_plot.index_name
        self.df = df
        self.excluded_isotopes = []  # Keep track of excluded isotopes
        self.app = dash.Dash(__name__)
        self.app.layout = html.Div([
            dcc.Graph(id='interactive-scatter', figure=self.fig, style={'height': '100vh'})
        ], style={'margin': 0})
        self.setup_callbacks()

    def setup_callbacks(self):
        @self.app.callback(
            Output('interactive-scatter', 'figure'),
            Input('interactive-scatter', 'restyleData'),
            State('interactive-scatter', 'figure')
        )
        def update_figure_on_legend_click(restyleData, current_figure_state):
            if restyleData and 'visible' in restyleData[0]:
                current_fig = go.Figure(current_figure_state)

                # Get the index of the clicked trace
                clicked_trace_index = restyleData[1][0]

                # Get the name of the clicked trace
                clicked_trace_name = current_fig.data[clicked_trace_index].name

                # Update excluded isotopes based on the clicked trace
                if restyleData[0]['visible'][0] == 'legendonly' and clicked_trace_name not in self.excluded_isotopes:
                    self.excluded_isotopes.append(clicked_trace_name)
                elif restyleData[0]['visible'][0] == True and clicked_trace_name in self.excluded_isotopes:
                    self.excluded_isotopes.remove(clicked_trace_name)

                # Update DataFrame based on excluded isotopes
                updated_df = self.df.copy()
                updated_df = updated_df[~updated_df['Isotope'].isin(self.excluded_isotopes)]

                # Recalculate the regression and summary statistics
                self.add_regression_and_stats(updated_df)

                # Update trace visibility based on excluded isotopes
                for trace in self.fig.data:
                    if trace.name in self.excluded_isotopes:
                        trace.visible = 'legendonly'
                    else:
                        trace.visible = True

                return self.fig

            return dash.no_update

        @self.app.server.route('/shutdown', methods=['POST'])
        def shutdown():
            os.kill(os.getpid(), signal.SIGINT)  # Send the SIGINT signal to the current process
            return 'Server shutting down...'

    def update_figure_on_legend_click(self, restyleData, current_figure_state):
        if restyleData and 'visible' in restyleData[0]:
            current_fig = go.Figure(current_figure_state)

            # Get the index of the clicked trace
            clicked_trace_index = restyleData[1][0]

            # Get the name of the clicked trace
            clicked_trace_name = current_fig.data[clicked_trace_index].name

            # Update excluded isotopes based on the clicked trace
            if restyleData[0]['visible'][0] == 'legendonly' and clicked_trace_name not in self.excluded_isotopes:
                self.excluded_isotopes.append(clicked_trace_name)
            elif restyleData[0]['visible'][0] == True and clicked_trace_name in self.excluded_isotopes:
                self.excluded_isotopes.remove(clicked_trace_name)

            # Update DataFrame based on excluded isotopes
            updated_df = self.df.copy()
            updated_df = updated_df[~updated_df['Isotope'].isin(self.excluded_isotopes)]

            # Create a new _InteractiveScatterPlotter instance with the updated data
            updated_plotter = InteractiveScatterPlotter(self.index_name, self.interactive_scatter_plot.nested)
            updated_plotter.create_plot(self.interactive_scatter_plot._create_scatter_data(updated_df), updated_df['Isotope'].unique(), updated_df['Reaction'].unique() if 'Reaction' in updated_df.columns else [])

            # Update the current figure with the new traces and layout
            current_fig.data = updated_plotter.fig.data
            current_fig.layout = updated_plotter.fig.layout

            # Update trace visibility based on excluded isotopes
            for trace in current_fig.data:
                if trace.name in self.excluded_isotopes:
                    trace.visible = 'legendonly'
                else:
                    trace.visible = True

            return current_fig

        return dash.no_update

    def show(self):
        port = find_free_port()
        # Function to open the browser
        def open_browser():
            if not os.environ.get("WERKZEUG_RUN_MAIN"):
                print(f"Now running at http://localhost:{port}/")
                webbrowser.open(f"http://localhost:{port}/")

        # Silence the Flask development server logging
        log = open(os.devnull, 'w')
        # sys.stdout = log
        sys.stderr = log

        # Disable Flask development server warning
        os.environ['FLASK_ENV'] = 'development'

        # JavaScript code to detect when the tab or window is closed
        self.app.index_string = '''
        <!DOCTYPE html>
        <html>
            <head>
                {%metas%}
                <title>{%title%}</title>
                {%favicon%}
                {%css%}
            </head>
            <body style="margin: 0;">
                {%app_entry%}
                <footer>
                    {%config%}
                    {%scripts%}
                    <script type="text/javascript">
                        window.addEventListener("beforeunload", function (e) {
                            var xhr = new XMLHttpRequest();
                            xhr.open("POST", "/shutdown", false);
                            xhr.send();
                        });
                    </script>
                    {%renderer%}
                </footer>
            </body>
        </html>
        '''

        # Timer to open the browser shortly after the server starts
        threading.Timer(1, open_browser).start()

        self.app.run_server(debug=False, host='localhost', port=port)

    def save_state(self, filename=None):
        state = {
            'fig': self.fig.to_dict(),
            'df': self.df.to_dict(),
            'excluded_isotopes': self.excluded_isotopes,
            'index_name': self.index_name,
            'nested': self.interactive_scatter_plot.nested
        }
        if filename is None:
            return state
        else:
            with open(filename, 'wb') as f:
                pickle.dump(state, f)

    @classmethod
    def load_state(cls, filename=None, data_dict=None):
        if filename is None and data_dict is None:
            raise ValueError("Either a filename or a data dictionary must be provided")
        if filename is not None:
            with open(filename, 'rb') as f:
                state = pickle.load(f)
        else:
            state = data_dict

        # Recreate the _InteractiveScatterPlotter instance from the saved state
        fig = go.Figure(state['fig'])
        index_name = state['index_name']
        nested = state['nested']
        interactive_scatter_plot = InteractiveScatterPlotter(index_name, nested)
        interactive_scatter_plot.fig = fig

        # Recreate the InteractiveScatterLegend instance from the saved state
        instance = cls(interactive_scatter_plot, pd.DataFrame.from_dict(state['df']))
        instance.excluded_isotopes = state['excluded_isotopes']

        # Update trace visibility based on excluded isotopes
        for trace in instance.fig.data:
            if trace.name in instance.excluded_isotopes:
                trace.visible = 'legendonly'
            else:
                trace.visible = True

        return instance

    def write_html(self, filename):
        # Utilize Plotly's write_html to save the current state of the figure
        self.fig.write_html(filename)



import socket
from tsunami_ip_utils.utils import filter_redundant_reactions, isotope_reaction_list_to_nested_dict
import numpy as np
from typing import Dict, List
from uncertainties import unumpy
from pathlib import Path

def find_free_port():
    """Finds a free port on localhost for running a Flask server."""
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.bind(("", 0))  # Let the OS pick an available port
    port = s.getsockname()[1]
    s.close()
    return port

def determine_plot_type(contributions, plot_redundant_reactions):
    """Determines whether the contributions are nuclide-wise or nuclide-reaction-wise and whether to plot redundant
    reactions or not
    
    Parameters
    ----------
    - contributions: list of dict, list of dictionaries containing the contributions to the similarity parameter for each
        nuclide or nuclide-reaction pair
    - plot_redundant_reactions: bool, whether to plot redundant reactions (or irrelevant reactions) when considering
        nuclide-reaction-wise contributions
        
    Returns
    -------
    - contributions: dict, contributions to the similarity parameter keyed by isotope then by reaction type"""
    if 'reaction_type' in contributions[0]: # Nuclide-reaction-wise contributions
        nested_plot = True # Nested plot by nuclide then by reaction type

        # Create a dictionary of contributions keyed by isotope then by reaction type
        contributions = isotope_reaction_list_to_nested_dict(contributions, 'contribution')

        # If viewing nuclide-reaction wise contributions, it's important (at least for the visualizations in this function)
        # that if viewing the true contributions to the nuclide total, that redundant interactions (e.g. capture and fission
        # + (n, g)) and irrelevant interactions (e.g. chi and nubar) are not plotted.

        if not plot_redundant_reactions:
            # Remove redundant interactions
            contributions = filter_redundant_reactions(contributions)
    else: # Nuclide-wise contributions
        nested_plot = False
        contributions = { contribution['isotope']: contribution['contribution'] for contribution in contributions }

    return contributions, nested_plot

def generate_plot_objects_from_array_contributions(contributions: Dict[ str, List[ unumpy.uarray ] ], integral_index_name: str, 
                                                   **kwargs: dict) -> np.ndarray:
    """Generate a matrix of plot objects (for creating a matrix plot) for the given contributions to an arbitrary integral index.
    This is valid for plots of %Δk/k, E contributions, c_k contributions, etc..
    
    Parameters
    ----------
    contributions
        Dictionary of a list of contributions to the integral index keyed by application or experiment.
    integral_index_name
        Name of the integral index being plotted.
    kwargs
        Additional keyword arguments. The following are supported:
        * diagonal_type : str
            Type of plot to create on the diagonal. Default is ``'interactive_pie'`` which creates an interactive
            pie chart.
        * interactive_contribution_legend : bool
            Whether to make the legend interactive for the contribution plots. Default is ``True``.
        * interactive_correlation_legend : bool
            Whether to make the legend interactive for the correlation plots. Default is ``True``.
            
    Returns
    -------
        2D numpy array of plot objects to be plotted with the matrix plot function."""
    from tsunami_ip_utils.viz import contribution_plot, correlation_plot # Import here to avoid circular import
    
    # Get options for legend interactivity and the diagonal plot type if supplied
    diagonal_type = kwargs.get('diagonal_type', 'interactive_pie')
    interactive_correlation_legend = kwargs.get('interactive_correlation_legend', True)
    interactive_contribution_legend = kwargs.get('interactive_contribution_legend', True)
    
    num_applications = len(contributions['application'])
    num_experiments = len(contributions['experiment'])

    # Construct plot matrix
    plot_objects_array = np.empty( ( num_applications, num_experiments ), dtype=object )

    for application_index in range(num_applications):
        for experiment_index in range(num_experiments):
            if experiment_index == application_index:
                # On the diagonal, make a contribution plot, as a correlation plot is not useful when comparing the same
                # application and experiment
                plot_objects_array[application_index, experiment_index] = \
                contribution_plot(
                    contributions['application'][application_index],
                    plot_type=diagonal_type,
                    integral_index_name=integral_index_name,
                    interactive_legend=interactive_contribution_legend,     
                )
            else:
                plot_objects_array[application_index, experiment_index] = \
                correlation_plot(
                    contributions['application'][application_index], 
                    contributions['experiment'][experiment_index], 
                    plot_type=diagonal_type,
                    integral_index_name=integral_index_name, 
                    plot_redundant_reactions=True, 
                    interactive_legend=interactive_correlation_legend
                )

    return plot_objects_array

def generate_plot_objects_array_from_perturbations(points_array: np.ndarray) -> np.ndarray:
    """Generate a matrix of plot objects (for creating a matrix plot) for the given contributions to an arbitrary integral index.
    This is valid for plots of %Δk/k, E contributions, c_k contributions, etc..
    
    Parameters
    ----------
    points_array
        Array of points generated from the perturbation test.
            
    Returns
    -------
        2D numpy array of plot objects to be plotted with the matrix plot function."""
    from tsunami_ip_utils.viz import perturbation_plot # Import here to avoid circular import
    
    # Construct plot matrix
    plot_objects_array = np.empty_like(points_array, dtype=object)

    for i, row in enumerate(points_array):
        for j, _ in enumerate(row):
            plot_objects_array[i, j] = perturbation_plot(points_array[i, j])

    return plot_objects_array
    



import dash
from dash import dcc, html
from dash.dependencies import Input, Output, State
import numpy as np
from pathlib import Path
import plotly.graph_objects as go
from .scatter_plot import InteractiveScatterLegend
from .pie_plot import InteractivePieLegend
import webbrowser
import os
import sys
import threading
from .plot_utils import find_free_port
import pickle

# Style constants
GRAPH_STYLE = {
    'flex': '1',
    'minWidth': '800px',
    'height': '500px',
    'padding': '10px',
    'borderRight': '1px solid black',
    'borderBottom': '1px solid black',
    'borderTop': '0px',
    'borderLeft': '0px'
}

def create_app(external_stylesheets):
    return dash.Dash(__name__, external_stylesheets=external_stylesheets)

def create_column_headers(num_cols):
    return [html.Div(
        f'Application {i+1}', 
        style={
            'flex': '1', 
            'minWidth': '800px', 
            'textAlign': 'center', 
            'padding': '10px', 
            'borderRight': '1px solid black', 
            'borderBottom': '1px solid black', 
            'display': 'flex', 
            'alignItems': 'center', 
            'justifyContent': 'center'
        }
    ) for i in range(num_cols)]

def create_row_label(i):
    return html.Div(
        html.Span(
            f'Experiment {i+1}',
            style={
                'display': 'block',
                'overflow': 'visible',
                'transform': 'rotate(-90deg)',
                'transformOrigin': 'center',
                'whiteSpace': 'nowrap',
            }
        ), 
        style={
            'flex': 'none',
            'width': '50px', 
            'textAlign': 'center', 
            'marginRight': '0', 
            'padding': '10px', 
            'borderRight': '1px solid black', 
            'borderBottom': '1px solid black', 
            'display': 'flex', 
            'alignItems': 'center', 
            'justifyContent': 'center'
        }
    )

def create_plot_element(i, j, plot_object):
    if isinstance(plot_object, InteractiveScatterLegend):
        graph_id = f"interactive-scatter-{i}-{j}"
        return dcc.Graph(id=graph_id, figure=plot_object.fig, style=GRAPH_STYLE)
    elif isinstance(plot_object, InteractivePieLegend):
        with plot_object.app.test_client() as client:
            response = client.get('/')
            html_content = response.data.decode('utf-8')
            return html.Iframe(srcDoc=html_content, style=GRAPH_STYLE)
    else:
        return dcc.Graph(figure=plot_object, style=GRAPH_STYLE)

def create_update_figure_callback(app, graph_id, app_instance):
    @app.callback(
        Output(graph_id, 'figure'),
        Input(graph_id, 'restyleData'),
        State(graph_id, 'figure')
    )
    def update_figure_on_legend_click(restyleData, current_figure_state):
        if restyleData and 'visible' in restyleData[0]:
            current_fig = go.Figure(current_figure_state)

            # Get the index of the clicked trace
            clicked_trace_index = restyleData[1][0]

            # Get the name of the clicked trace
            clicked_trace_name = current_fig.data[clicked_trace_index].name

            # Update excluded isotopes based on the clicked trace
            if restyleData[0]['visible'][0] == 'legendonly' and clicked_trace_name not in app_instance.excluded_isotopes:
                app_instance.excluded_isotopes.append(clicked_trace_name)
            elif restyleData[0]['visible'][0] == True and clicked_trace_name in app_instance.excluded_isotopes:
                app_instance.excluded_isotopes.remove(clicked_trace_name)

            # Update DataFrame based on excluded isotopes
            updated_df = app_instance.df.copy()
            updated_df = updated_df[~updated_df['Isotope'].isin(app_instance.excluded_isotopes)]

            # Recalculate the regression and summary statistics
            app_instance.add_regression_and_stats(updated_df)

            # Update trace visibility based on excluded isotopes
            for trace in app_instance.fig.data:
                if trace.name in app_instance.excluded_isotopes:
                    trace.visible = 'legendonly'
                else:
                    trace.visible = True

            return app_instance.fig

        return dash.no_update

def generate_layout(app, rows):
    app.layout = html.Div([
        html.H1("Matrix of Plots", style={'textAlign': 'center', 'marginLeft': '121px'}),
        html.Div(rows, style={'display': 'flex', 'flexDirection': 'column', 'width': '100%', 'overflowX': 'auto'}),
        html.Script("""
        window.addEventListener('resize', function() {
            const graphs = Array.from(document.querySelectorAll('.js-plotly-plot'));
            graphs.forEach(graph => {
                Plotly.Plots.resize(graph);
            });
        });
        """)
    ])

class InteractiveMatrixPlot:
    def __init__(self, app, plot_objects_array):
        self.app = app
        self.plot_objects_array = plot_objects_array
    
    def open_browser(self, port):
        print(f"Now running at http://localhost:{port}/")
        webbrowser.open(f"http://localhost:{port}/")
        pass

    def show(self, open_browser=True, silent=False):
        """Start the Flask server and open the browser to display the interactive sunburst chart
        
        Parameters
        ----------
        - open_browser: bool, whether to open the browser automatically to display the chart
        - silent: bool, whether to suppress Flask's startup and runtime messages"""
        # Suppress Flask's startup and runtime messages by redirecting them to dev null
        log = open(os.devnull, 'w')
        # sys.stdout = log
        sys.stderr = log

        port = find_free_port()
        if open_browser:
            threading.Timer(1, self.open_browser(port)).start()
        self.app.run(host='localhost', port=port)

    def write_html(self, filename=None):
        with self.app.test_client() as client:
            response = client.get('/')
            html_content = response.data.decode('utf-8')
            if filename is None:
                return html_content
            else:
                with open(filename, 'w') as f:
                    f.write(html_content)
    
    def save_state(self, filename):
        # Serialize interactive plots in the plot objects array
        self.plot_types = np.empty_like(self.plot_objects_array, dtype=object)
        for i, row in enumerate(self.plot_objects_array):
            for j, plot_object in enumerate(row):
                if isinstance(plot_object, InteractiveScatterLegend):
                    self.plot_objects_array[i,j] = plot_object.save_state()
                    self.plot_types[i,j] = "InteractiveScatterLegend"
                elif isinstance(plot_object, InteractivePieLegend):
                    self.plot_objects_array[i,j] = plot_object.save_state()
                    self.plot_types[i,j] = "InteractivePieLegend"

        with open(filename, 'wb') as f:
            pickle.dump( ( self.plot_objects_array, self.plot_types ) , f)

    @classmethod
    def load_state(self, filename):
        with open(filename, 'rb') as f:
            plot_objects_array, plot_types = pickle.load(f)
            # Reserialize interactive scatter legends
            for i, row in enumerate(plot_objects_array):
                for j, plot_object in enumerate(row):
                    if plot_types[i,j] == "InteractiveScatterLegend":
                        plot_objects_array[i,j] = InteractiveScatterLegend.load_state(data_dict=plot_object)
                    elif plot_types[i,j] == "InteractivePieLegend":
                        plot_objects_array[i,j] = InteractivePieLegend.load_state(data_dict=plot_object)

        return interactive_matrix_plot(plot_objects_array)
    

def load_interactive_matrix_plot(filename):
    """Loads an interactive matrix plot from a saved state pickle file. This function is purely for convenience and is a
    wrapper of the InteractiveScatterLegend.load_state method"""
    return InteractiveMatrixPlot.load_state(filename)


def interactive_matrix_plot(plot_objects_array: np.ndarray):
    current_directory = Path(__file__).parent
    external_stylesheets = [str(current_directory / 'css' / 'matrix_plot.css')]
    app = create_app(external_stylesheets)

    num_rows = plot_objects_array.shape[0]
    num_cols = plot_objects_array.shape[1]

    column_headers = create_column_headers(num_cols)
    header_row = html.Div([html.Div('', style={'flex': 'none', 'width': '71px', 'borderBottom': '1px solid black'})] + column_headers, style={'display': 'flex'})

    rows = [header_row]
    for i in range(num_rows):
        row = [create_row_label(i)]
        for j in range(num_cols):
            plot_object = plot_objects_array[i, j]
            plot_element = create_plot_element(i, j, plot_object) if plot_object else html.Div('Plot not available', style=GRAPH_STYLE)
            row.append(plot_element)
            if isinstance(plot_object, InteractiveScatterLegend):
                create_update_figure_callback(app, f"interactive-scatter-{i}-{j}", plot_object)
        rows.append(html.Div(row, style={'display': 'flex'}))

    generate_layout(app, rows)
    return InteractiveMatrixPlot(app, plot_objects_array)


from uncertainties import unumpy
import numpy as np

def _unit_vector_uncertainty_propagation(vector: unumpy.uarray) -> np.ndarray:
    """Performs error propagation for the components of a vector :math:`\\boldsymbol{v}` that is normalized to a unit vector 
    via :math:`\\hat{\\boldsymbol{v}} = \\frac{\\boldsymbol{v}}{\\lVert\\boldsymbol{v}\\rVert}`. To calculate the uncertainty 
    of :math:`\\hat{\\boldsymbol{v}}\\equiv \\boldsymbol{u}`, recall from linear error propagation theory 
    (see `this <https://en.wikipedia.org/wiki/Propagation_of_uncertainty#Simplification>`_ equation)

    .. math::
        \\sigma_{u_i} = \\sqrt{ \\sum_j \\left( \\frac{\\partial u_i}{\\partial v_j} \\sigma_{v_j}  \\right)^2 }

    since
    
    .. math::
        u_i = \\frac{v_i}{\\sqrt{\\sum_j v_j^2}}
        
    we have for :math:`i=j`:

    .. math::
        \\frac{\\partial u_i}{\\partial v_j} &= 
        \\frac{1}{\\sqrt{\\sum_j v_j^2}} - \\frac{1}{2}\\frac{v_i\\cdot 2v_i}{\\left(\\sum_j v_j^2\\right)^{\\frac{3}{2}}}\\\\
        &=\\frac{1}{\\lVert \\boldsymbol{v}\\rVert} - \\frac{v_i^2}{\\lVert \\boldsymbol{v}\\rVert^3}\\\\
        &=\\frac{\\lVert \\boldsymbol{v}\\rVert^2 - v_i^2}{\\lVert \\boldsymbol{v}\\rVert^3}

    For :math:`i\\neq j`:

    .. math::
        \\frac{\\partial u_i}{\\partial v_j} = -\\frac{v_iv_j}{\\lVert\\boldsymbol{v}\\rVert^3}

    Examples
    --------
    >>> unit_vector_uncertainty_propagation(unumpy.uarray([1, 2, 3], [0.1, 0.2, 0.3]))
    array([0.03113499, 0.05150788, 0.03711537])

    >>> unit_vector_uncertainty_propagation(unumpy.uarray([0, 0, 0], [0.1, 0.2, 0.3]))
    array([0., 0., 0.])

    >>> unit_vector_uncertainty_propagation(unumpy.uarray([0, 0, 0], [0.0, 0.0, 0.0]))
    array([0., 0., 0.])
    
    Parameters
    ----------
    vector
        The vector :math:`\\boldsymbol{v}` (with uncertainties) whose normalization the error propagation is
        calculated for.
    
    Returns
    -------
        Uncertainties of the unit vector components."""

    # Calculate norm of the vector
    vector_norm = np.sqrt(np.sum(unumpy.nominal_values(vector)**2))

    # Extract the uncertainties of the vector components
    vector_uncertainties = unumpy.std_devs(vector)

    # Compute the derivative matrix for uncertainty propagation
    if vector_norm != 0:
        derivative_matrix = -unumpy.nominal_values(vector)[:, np.newaxis] * unumpy.nominal_values(vector)[np.newaxis, :] / vector_norm**3
        np.fill_diagonal(derivative_matrix, (vector_norm**2 - unumpy.nominal_values(vector)**2) / vector_norm**3)
    else:
        derivative_matrix = np.zeros((len(vector), len(vector)))

    # Calculate the uncertainties of the unit vector components
    unit_vector_uncertainties = np.sqrt(np.sum((derivative_matrix**2 * vector_uncertainties**2), axis=1))

    # Return the unit vector with uncertainties
    return unit_vector_uncertainties

def _dot_product_uncertainty_propagation(vect1: unumpy.uarray, vect2: unumpy.uarray) -> float:
    """Calculates the uncertainty in the dot product of two vectors with uncertainties
    
    Parameters
    ----------
    vect1
        First vector in dot product.
    vect2
        Second vector in dot product.

    Returns
    -------
        Uncertainty in the dot product of the two vectors."""

    dot_product_uncertainty = 0
    for i in range(len(vect1)):
        if vect1[i].n == 0 or vect2[i].n == 0:
            product_uncertainty = 0
        else:
            product_uncertainty = vect1[i].n * vect2[i].n * np.sqrt((vect1[i].s/vect1[i].n)**2 + \
                                                                    (vect2[i].s/vect2[i].n)**2)
        dot_product_uncertainty += product_uncertainty**2
    dot_product_uncertainty = np.sqrt(dot_product_uncertainty)

    return dot_product_uncertainty



import doctest
import os
import importlib
import pkgutil

def find_and_run_doctests(package_name):
    package = importlib.import_module(package_name)
    print(f"Testing package: {package_name}")
    print('-'*len(f"Testing package: {package_name}"))
    errors = 0
    failures = 0

    for importer, modname, ispkg in pkgutil.walk_packages(package.__path__, package.__name__ + '.'):
        try:
            module = importlib.import_module(modname)
            print(f"Testing module: {modname}")
            result = doctest.testmod(module)
            failures += result.failed
            if result.failed:
                print(f"Failed: {result.failed}, Attempted: {result.attempted}")
        except Exception as e:
            print(f"Failed to test module {modname}: {str(e)}")

    print(f"\nFinished testing. Failures: {failures}")

# Replace 'mypackage' with your actual package name.
if __name__ == "__main__":
    find_and_run_doctests('tsunami_ip_utils')


import re

def isotope_reaction_list_to_nested_dict(isotope_reaction_list, field_of_interest):
    """Converts a list of dictionaries containing isotope-reaction pairs (and some other key that represents a value of
    interest, e.g. an sdf profile or a contribution) to a nested dictionary
    
    Parameters
    ----------
    - isotope_reaction_list: list of dict, list of dictionaries containing isotope-reaction pairs and some other key
    - field_of_interest: str, the key in the dictionary that represents the value of interest

    Returns
    -------
    - nested_dict: dict, nested dictionary containing the isotope-reaction pairs and the value of interest"""

    isotope_reaction_dict = {}

    def get_atomic_number(isotope):
        return int(re.findall(r'\d+', isotope)[0])
    
    # Sort isotopes by atomic number so plots will have similar colors across different calls
    all_isotopes = list(set([isotope_reaction['isotope'] for isotope_reaction in isotope_reaction_list]))
    all_isotopes.sort(key=get_atomic_number)
    isotope_reaction_dict = { isotope: {} for isotope in all_isotopes }

    for isotope_reaction in isotope_reaction_list:
        isotope = isotope_reaction['isotope']
        reaction = isotope_reaction['reaction_type']
        value = isotope_reaction[field_of_interest]

        isotope_reaction_dict[isotope][reaction] = value

    return isotope_reaction_dict

def filter_redundant_reactions(data_dict, redundant_reactions=['chi', 'capture', 'nubar', 'total']):
    """Filters out redundant reactions from a nested isotope-reaction dictionary
    
    Parameters
    ----------
    - data_dict: dict, nested dictionary containing isotope-reaction pairs
    - redundant_reactions: list of str, list of reactions to filter out"""
    return { isotope: { reaction: data_dict[isotope][reaction] for reaction in data_dict[isotope] \
                        if reaction not in redundant_reactions } for isotope in data_dict }

def filter_by_nuclie_reaction_dict(data_dict, nuclide_reactions):
    """Filters out isotopes that are not in the nuclide_reactions dictionary
    
    Parameters
    ----------
    - data_dict: dict, nested dictionary containing isotope-reaction pairs
    - nuclide_reactions: dict, dictionary containing isotopes and their reactions"""
    return {nuclide: {reaction: xs for reaction, xs in reactions.items() if reaction in nuclide_reactions[nuclide]} \
                        for nuclide, reactions in data_dict.items() if nuclide in nuclide_reactions.keys()}


import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import plotly.express as px
from .base_plotter import Plotter
from plotly.subplots import make_subplots
import os, sys, signal
import threading
import webbrowser
from flask import Flask, render_template_string
import uuid
from .plot_utils import find_free_port
import pickle


class PiePlotter(Plotter):
    def __init__(self, integral_index_name, plot_redudant=False, **kwargs):
        self.index_name = integral_index_name
        self.plot_redundant = plot_redudant
    
    def create_plot(self, contributions, nested):
        self.nested = nested
        self.fig, self.axs = plt.subplots()
        if nested:
            self.nested_pie_chart(contributions)
        else:
            self.pie_chart(contributions)

        self.style()

    def add_to_subplot(self, fig, position):
        return fig.add_subplot(position, sharex=self.ax, sharey=self.ax)

    def get_plot(self):
        return self.fig, self.axs

    def nested_pie_chart(self, contributions):
        # Create a nested ring chart
        num_reactions = len(next(iter(contributions.values())))
        nuclide_colors = plt.get_cmap('rainbow')(np.linspace(0, 1, len(contributions.keys())))
        nuclide_totals = { nuclide: sum(contribution.n for contribution in contributions[nuclide].values()) \
                        for nuclide in contributions }
        nuclide_labels = list(nuclide_totals.keys())

        # Now, deal with negative values

        nuclides_with_opposite_sign_contributions = []
        for nuclide, contribution in contributions.items():
            contribution_values = [contribution[reaction].n for reaction in contribution]
            if not (all(v >= 0 for v in contribution_values) or all(v <= 0 for v in contribution_values)):
                nuclides_with_opposite_sign_contributions.append(nuclide)
            
        # For nuclides with opposite sign contributions, we distinguish the positive and negative contributions
        # by coloring some of the inner ring a lighter color to indicate the negative contributions in the outer ring
        wedge_widths = list(nuclide_totals.values())
        inner_wedge_hatches = [None] * len(wedge_widths)

        def blend_colors(color1, color2, alpha):
            return np.array( [ alpha * c1 + (1 - alpha) * c2 for c1, c2 in zip(color1, color2 ) ] )

        if len(nuclides_with_opposite_sign_contributions) > 0:
            for nuclide in nuclides_with_opposite_sign_contributions:
                # First, determine the fraction of the contributions that are opposite (in sign) to the total
                total_sign = np.sign(nuclide_totals[nuclide])
                
                # Now, we want to plot the "lost" wedge width in white, i.e. the width lost from cancellations between the
                # positive and negative contributions. This will be colored a lighter color. The absolute sum of the
                # contributions represents the wedge width if there were no cancellations, so the total wedge width
                # minus the absolute sum of the contributions is "lost" wedge width.

                absolute_sum_of_contributions = sum(np.abs(contribution.n) for contribution in contributions[nuclide].values())
                
                # NOTE the sign function is needed to handle the case when the nuclide total is negative
                lost_wedge_width = absolute_sum_of_contributions - total_sign * nuclide_totals[nuclide]

                # Now, insert the lost wedge width into the wedge widths list right after the nuclide
                nuclide_index = list(nuclide_totals.keys()).index(nuclide)
                wedge_widths.insert(nuclide_index + 1, lost_wedge_width)
                nuclide_labels.insert(nuclide_index + 1, '')
                
                # The color of the lost wedge width will be a blend of the nuclide color and white
                white_color = np.array([1, 1, 1, 1])
                opacity = 0.8
                blended_color = blend_colors(white_color, nuclide_colors[nuclide_index], opacity)
                nuclide_colors = np.insert(nuclide_colors, nuclide_index + 1, blended_color, axis=0)
                
                # Add hatches to the negative total sum wedge
                if nuclide_totals[nuclide] < 0:
                    inner_wedge_hatches[nuclide_index] = '//'

        # Now make everything positive for the pie chart
        wedge_widths = np.abs(wedge_widths)

        # Plot the inner ring for nuclide totals
        inner_ring, _ = self.axs.pie(wedge_widths, radius=0.7, labels=nuclide_labels, \
                                colors=nuclide_colors, labeldistance=0.6, textprops={'fontsize': 8}, \
                                    wedgeprops=dict(width=0.3, edgecolor='w'))

        # Add hatches to the negative total sum wedges
        for wedge, hatch in zip(inner_ring, inner_wedge_hatches):
            if hatch:
                wedge.set_hatch(hatch)

        # Get colors for reactions from the "rainbow" colormap
        reaction_colors = plt.get_cmap('Set1')(np.linspace(0, 1, num_reactions))

        # Plot the outer ring for reaction-specific contributions
        outer_labels = []
        outer_colors = []
        outer_sizes = []
        outer_hatches = []
        for i, (nuclide, reactions) in enumerate(contributions.items()):
            for j, (reaction, contribution) in enumerate(list(reactions.items())):
                outer_labels.append(f"{nuclide} - {reaction}")
                
                outer_colors.append(reaction_colors[j])
                outer_sizes.append(np.abs(contribution.n))
                
                if contribution.n < 0:
                    outer_hatches.append('//')
                else:
                    outer_hatches.append(None)

        outer_ring, _ = self.axs.pie(outer_sizes, radius=1, labels=outer_labels, labeldistance=0.9, colors=outer_colors, \
                textprops={'fontsize': 6}, startangle=inner_ring[0].theta1, counterclock=True, \
                    wedgeprops=dict(width=0.3, edgecolor='w'))

        # Add hatches to the negative contribution wedges
        for wedge, hatch in zip(outer_ring, outer_hatches):
            if hatch:
                wedge.set_hatch(hatch)
        
    def pie_chart(self, contributions):
        labels = list(contributions.keys())
        values = [abs(contributions[key].n) for key in labels]

        # Determining hatching patterns: empty string for positive, cross-hatch for negative
        hatches = ['//' if contributions[key].n < 0 else '' for key in labels]

        # Creating the pie chart
        wedges, _ = self.axs.pie(values, labels=labels, startangle=90)

        # Applying hatching patterns to the wedges
        for wedge, hatch in zip(wedges, hatches):
            wedge.set_hatch(hatch)

    def style(self):
        if self.plot_redundant and self.nested:
            title_text = f'Contributions to {self.index_name} (including redundant/irrelvant reactions)'
        else:
            title_text = f'Contributions to {self.index_name}'
        self.axs.grid(True, which='both', axis='y', color='gray', linestyle='-', linewidth=0.5)
        self.axs.set_title(title_text)


class InteractivePiePlotter(Plotter):
    def __init__(self, integral_index_name, plot_redundant=False, **kwargs):
        # Check if the user wants an interactive legend
        if 'interactive_legend' in kwargs.keys():
            self.interactive_legend = kwargs['interactive_legend']
        else:
            self.interactive_legend = True
        
        self.index_name = integral_index_name
        self.plot_redundant = plot_redundant

    def create_plot(self, contributions, nested=True):
        self.fig = make_subplots()

        # Prepare data for the sunburst chart
        self.nested = nested
        if nested:
            df = self._create_nested_sunburst_data(contributions)
        else:
            df = self._create_sunburst_data(contributions)
        
        # Create a sunburst chart
        self.fig = px.sunburst(
            data_frame=df,
            names='labels',
            parents='parents',
            ids='ids',
            values='normalized_values',
            custom_data=['values', 'uncertainties']
        )

        # Update hovertemplate with correct syntax
        self.fig.update_traces(
            hovertemplate=(
                "<b>%{label}</b><br>"
                "Value: %{customdata[0]:1.4E} +/- %{customdata[1]:1.4E}"  # Corrected format specifiers
                "<extra></extra>"  # This hides the trace info
            )
        )

        # Now style the plot
        self.style()

        self.fig.update_layout(
            autosize=True,
            width=None,  # Removes fixed width
            height=None,  # Removes fixed height
            margin=dict(l=5, r=5, t=30, b=5)
        )

        if self.interactive_legend:
            self.fig = InteractivePieLegend(self.fig, df)


    
    def add_to_subplot(self, fig, position):
        if self.interactive_legend:
            raise ValueError("Interactive legends are not supported when adding to a subplot")
        else:
            for trace in self.fig.data:
                fig.add_trace(trace, row=position[0], col=position[1])
            return fig

    def get_plot(self):
        return self.fig

    def _create_sunburst_data(self, contributions):
        data = {
            'labels': [], 
            'ids': [], 
            'parents': [], 
            'values': [], 
            'uncertainties': [],
            'normalized_values': [],
            'nuclide': []
        }

        abs_sum_of_nuclide_totals = sum( abs(contribution.n) for contribution in contributions.values())

        for nuclide, nuclide_total in contributions.items():
            # Caclulate the nuclide total, and the positive and negative contributions
            norm_nuclide_total = abs(nuclide_total) / abs_sum_of_nuclide_totals

            # Add the nuclide as a parent
            data['labels'].append(nuclide)
            data['ids'].append(nuclide)
            data['parents'].append('')
            data['values'].append(nuclide_total.n)
            data['uncertainties'].append(nuclide_total.s)
            data['normalized_values'].append(norm_nuclide_total.n)
            data['nuclide'].append(nuclide)

        return pd.DataFrame(data)

    def _create_nested_sunburst_data(self, contributions):
        data = {
            'labels': [], 
            'ids': [], 
            'parents': [], 
            'values': [], 
            'uncertainties': [],
            'normalized_values': [],
            'nuclide': []
        }

        abs_sum_of_nuclide_totals = sum(sum(abs(contribution.n) for contribution in reactions.values()) \
                                    for reactions in contributions.values())

        for nuclide, reactions in contributions.items():
            # Caclulate the nuclide total, and the positive and negative contributions
            nuclide_total = sum(contribution for contribution in reactions.values())
            if abs_sum_of_nuclide_totals != 0:
                norm_nuclide_total = abs(nuclide_total) / abs_sum_of_nuclide_totals
            else:
                norm_nuclide_total = 0

            positive_contributions = { reaction: contribution for reaction, contribution in reactions.items() \
                                      if contribution.n >= 0 }
            negative_contributions = { reaction: contribution for reaction, contribution in reactions.items() \
                                      if contribution.n < 0 }
            positive_total = sum(contribution for contribution in positive_contributions.values())
            negative_total = sum(contribution for contribution in negative_contributions.values())

            # Add the nuclide as a parent
            data['labels'].append(nuclide)
            data['ids'].append(nuclide)
            data['parents'].append('')
            data['values'].append(nuclide_total.n)
            data['uncertainties'].append(nuclide_total.s)
            data['normalized_values'].append(norm_nuclide_total.n)
            data['nuclide'].append(nuclide)
    
            # --------------------------------------------------------
            # Add the positive and negative contributions as children
            # --------------------------------------------------------

            # Normalize the contributions by the absolute value of the nuclide total 
            absolute_sum = positive_total + abs(negative_total)
            if absolute_sum != 0:
                normalization_factor = abs(norm_nuclide_total) / absolute_sum
            else:
                normalization_factor = 0

            # Positive contributions
            if positive_total != 0:
                norm_positive_total = positive_total * normalization_factor
                data['labels'].append('Positive')
                data['ids'].append(f"{nuclide}-Positive")
                data['parents'].append(nuclide)
                data['values'].append(positive_total.n)
                data['uncertainties'].append(positive_total.s)
                data['normalized_values'].append( norm_positive_total.n )
                data['nuclide'].append(nuclide)
            else:
                norm_positive_total = 0

            # Negative contributions
            if negative_total != 0:
                norm_negative_total = abs(negative_total) * normalization_factor
                data['labels'].append('Negative')
                data['ids'].append(f"{nuclide}-Negative")
                data['parents'].append(nuclide)
                data['values'].append(negative_total.n)
                data['uncertainties'].append(negative_total.s)
                data['normalized_values'].append( norm_negative_total.n )
                data['nuclide'].append(nuclide)
            else:
                norm_negative_total = 0

            # -------------------------------
            # Add the reaction contributions
            # -------------------------------
            # NOTE: Plotly express apparently has issues dealing with small numbers, so unless the contribution is
            # multiplied by a sufficiently large scale factor, the data won't be displayed correctly
            scale_factor = 10000
            for reaction, contribution in positive_contributions.items():
                # Now normalize contributions so they sum to the "normalized_positive_total
                if positive_total != 0:
                    normalization_factor = norm_positive_total / positive_total
                else:
                    normalization_factor = 0
                norm_reaction_contribution = contribution.n * normalization_factor
                
                if contribution.n != 0:
                    data['labels'].append(reaction)
                    data['ids'].append(f"{nuclide}-{reaction}")
                    data['parents'].append(f"{nuclide}-Positive")
                    data['values'].append(contribution.n)
                    data['uncertainties'].append(contribution.s)
                    data['normalized_values'].append(scale_factor*norm_reaction_contribution.n)
                    data['nuclide'].append(nuclide)

            for reaction, contribution in negative_contributions.items():
                # Now normalize contributions so they sum to the "normalized_negative_total"
                normalization_factor = norm_negative_total / abs(negative_total)
                norm_reaction_contribution = abs(contribution.n) * normalization_factor

                if contribution.n != 0:
                    data['labels'].append(reaction)
                    data['ids'].append(f"{nuclide}-{reaction}")
                    data['parents'].append(f"{nuclide}-Negative")
                    data['values'].append(contribution.n)
                    data['uncertainties'].append(contribution.s)
                    data['normalized_values'].append(scale_factor*norm_reaction_contribution.n)
                    data['nuclide'].append(nuclide)


        return pd.DataFrame(data)

    def style(self):
        if self.plot_redundant and self.nested:
            title_text = f'Contributions to {self.index_name} (including redundant/irrelvant reactions)'
        else:
            title_text = f'Contributions to {self.index_name}'
        self.fig.update_layout(title_text=title_text, title_x=0.5)  # 'title_x=0.5' centers the title


class InteractivePieLegend:
    def __init__(self, fig, df):
        """Return a flask webapp that will display an interactive legend for the sunburst chart"""
        self.fig = fig
        self.df = df
        self.app = Flask(__name__)

        @self.app.route('/shutdown', methods=['POST'])
        def shutdown():
            os.kill(os.getpid(), signal.SIGINT)  # Send the SIGINT signal to the current process
            return 'Server shutting down...'

        @self.app.route('/')
        def show_sunburst():
            # Extract root nodes (nodes without parents)
            root_nodes = self.df[self.df['parents'] == '']

            # Generate a unique ID for the container
            container_id = f"container-{uuid.uuid4()}"

            # Generate legend HTML with a title
            legend_html = f'<div id="{container_id}-legend" style="border: 2px solid black; padding: 10px;"><h3 style="margin-top: 0; text-align: center;">Legend</h3>\n'
            for _, row in root_nodes.iterrows():
                legend_html += f'<div class="{container_id}-legend-item" style="cursor: pointer; margin-bottom: 5px;" data-target="{row["ids"]}">{row["ids"]}: {row["values"]:1.4E}</div>\n'
            legend_html += '</div>\n'

            # JavaScript for interactivity and shutdown
            script_html = f"""
            <script>
            window.addEventListener('beforeunload', (event) => {{
                navigator.sendBeacon('/shutdown');
            }});
            document.addEventListener('DOMContentLoaded', function () {{
                const legendItems = document.querySelectorAll('.{container_id}-legend-item');
                legendItems.forEach(item => {{
                    item.addEventListener('mouseenter', function() {{
                        const target = this.getAttribute('data-target');
                        const paths = document.querySelectorAll('path.surface');
                        paths.forEach(path => {{
                            const labelText = path.nextElementSibling ? path.nextElementSibling.textContent : "";
                            if (labelText.includes(target)) {{
                                path.style.opacity = 0.5; // Highlight by changing opacity
                            }}
                        }});
                    }});
                    item.addEventListener('mouseleave', function() {{
                        const paths = document.querySelectorAll('path.surface');
                        paths.forEach(path => {{
                            path.style.opacity = 1; // Reset opacity
                        }});
                    }});
                    item.addEventListener('click', function() {{
                        const target = this.getAttribute('data-target');
                        const paths = document.querySelectorAll('path.surface');
                        paths.forEach(path => {{
                            const labelText = path.nextElementSibling ? path.nextElementSibling.textContent : "";
                            if (labelText.includes(target)) {{
                                path.dispatchEvent(new MouseEvent('click', {{ 'view': window, 'bubbles': true, 'cancelable': true }}));
                            }}
                        }});
                    }});
                }});
                // Force Redraw/Reflow
                setTimeout(() => {{
                    window.dispatchEvent(new Event('resize'));
                }}, 100); // Delay may be adjusted based on actual rendering time
            }});
            </script>
            """

            # Save the chart with interactivity and layout adjustments
            fig_html = self.fig.to_html(full_html=False, include_plotlyjs='cdn')
            full_html = f"""
            <!DOCTYPE html>
            <html>
            <head>
            <title>Interactive Sunburst Chart</title>
            <style>
                #{container_id} {{
                    display: flex;
                    flex-direction: row; /* Align children horizontally */
                    height: 100%;
                    width: 100%; /* Ensure the container takes full width */
                    margin: 0;
                    font-family: Arial, sans-serif;
                }}
                #{container_id} > div {{
                    display: flex;
                    justify-content: space-between; /* Space out the chart and legend */
                    align-items: flex-start; /* Align items at the start of the cross axis */
                    width: 100%;
                    overflow: hidden; /* Hide overflow to prevent breaking layout */
                }}
                #{container_id}-chart {{
                    flex: 1 1 70%; /* Allow chart to grow and shrink but base at 70% width */
                    padding: 10px;
                }}
                #{container_id}-legend {{
                    flex: 0 1 30%; /* Start with 30% width but allow shrinking */
                    padding: 5px;
                    max-height: calc(100vh - 20px); /* Limit height to viewport height minus some margin */
                    overflow: auto; /* Scroll internally if content overflows */
                }}
            </style>
            </head>
            <body>
            <div id="{container_id}">
                <div>
                    <div id="{container_id}-chart">{fig_html}</div>
                    <div id="{container_id}-legend">{legend_html}</div>
                </div>
            </div>
            {script_html}
            </body>
            </html>
            """

            return render_template_string(full_html)
        
    def open_browser(self, port):
        print(f"Now running at http://localhost:{port}/")
        webbrowser.open(f"http://localhost:{port}/")
        pass

    def show(self, open_browser=True, silent=False):
        """Start the Flask server and open the browser to display the interactive sunburst chart
        
        Parameters
        ----------
        - open_browser: bool, whether to open the browser automatically to display the chart
        - silent: bool, whether to suppress Flask's startup and runtime messages"""
        # Suppress Flask's startup and runtime messages by redirecting them to dev null
        log = open(os.devnull, 'w')
        # sys.stdout = log
        sys.stderr = log

        port = find_free_port()
        if open_browser:
            threading.Timer(1, self.open_browser(port)).start()
        self.app.run(host='localhost', port=port)

    def serve(self):
        """Start the Flask server to display the interactive sunburst chart"""
        port = find_free_port()
        log = open(os.devnull, 'w')
        # sys.stdout = log
        # sys.stderr = log

        # Run the Flask application in a separate thread
        thread = threading.Thread(target=lambda: self.app.run(host='localhost', port=port))
        print(f"Now running at http://localhost:{port}/")
        thread.daemon = True  # This ensures thread exits when main program exits
        thread.start()

    def write_html(self, filename=None):
        with self.app.test_client() as client:
            response = client.get('/')
            html_content = response.data.decode('utf-8')
            if filename is None:
                return html_content
            else:
                with open(filename, 'w') as f:
                    f.write(html_content)

    def save_state(self, filename=None):
        state = {
            'fig': self.fig,
            'df': self.df,
        }
        if filename is None:
            return state
        else:
            with open(filename, 'wb') as f:
                pickle.dump(state, f)

    @classmethod
    def load_state(cls, filename=None, data_dict=None):
        if filename is None and data_dict is None:
            raise ValueError("Either a filename or a data dictionary must be provided")
        if filename is not None:
            with open(filename, 'rb') as f:
                state = pickle.load(f)
        else:
            state = data_dict
        fig = state['fig']
        df = state['df']
        instance = cls(fig, df)
        return instance


from .bar_plot import BarPlotter
from .pie_plot import PiePlotter, InteractivePiePlotter
from .scatter_plot import ScatterPlotter, InteractiveScatterPlotter, InteractivePerturbationScatterPlotter
from .matrix_plot import interactive_matrix_plot
from .plot_utils import determine_plot_type
from tsunami_ip_utils.integral_indices import _add_missing_reactions_and_nuclides
import numpy as np
from typing import List, Dict

def contribution_plot(contributions: List[Dict], plot_type: str='bar', integral_index_name: str='E', 
                      plot_redundant_reactions: bool=True, **kwargs: dict):
    """Plots the contributions to an arbitrary similarity parameter for a single experiment application pair
    
    Parameters
    ----------
    contributions
        list of dictionaries containing the contributions to the similarity parameter for each
        nuclide or nuclide-reaction pair
    plot_type
        type of plot to create. Default is 'bar' which creates a bar plot. Other option is 'pie' which creates
        a pie chart
    integral_index_name
        name of the integral index being plotted. Default is `'E'`
    plot_redundant_reactions
        whether to plot redundant reactions (or irrelevant reactions) when considering
        nuclide-reaction-wise contributions. Default is True
    kwargs
        additional keyword arguments to pass to the plotting function"""

    # Determine if the contributions are nuclide-wise or nuclide-reaction-wise
    contributions, nested_plot = determine_plot_type(contributions, plot_redundant_reactions)

    plotters = {
        'bar': BarPlotter(integral_index_name, plot_redundant_reactions, **kwargs),
        'pie': PiePlotter(integral_index_name, plot_redundant_reactions, **kwargs),
        'interactive_pie': InteractivePiePlotter(integral_index_name, plot_redundant_reactions, **kwargs)
    }
    
    # Get the requested plotter
    plotter = plotters.get(plot_type)
    if plotter is None:
        raise ValueError("Unsupported plot type")

    # Create the plot and style it
    plotter.create_plot(contributions, nested_plot)

    return plotter.get_plot()


def correlation_plot(application_contributions, experiment_contributions, plot_type='scatter', integral_index_name='E', \
                     plot_redundant_reactions=True, **kwargs):
    """Creates a correlation plot for a given application-experiment pair for which the contributions to the similarity
    parameter are given.
    
    Parameters
    ----------
    - application_contributions: list of dict, list of dictionaries containing the contributions to the similarity parameter
        for the application
    - experiment_contributions: list of dict, list of dictionaries containing the contributions to the similarity parameter
        for the experiment
    - plot_type: str, type of plot to create. Default is 'scatter' which creates a matplotlib scatter plot. Other options
        are interactive_scatter, which creates a Plotly scatter plot.
    - integral_index_name: str, name of the integral index being plotted. Default is 'E'
    - plot_redundant_reactions: bool, whether to plot redundant reactions (or irrelevant reactions) when considering
        nuclide-reaction-wise contributions. Default is True
        
    Returns
    -------
    - fig: matplotlib.figure.Figure, the figure containing the correlation plot"""

    # Determine if the contributions are nuclide-wise or nuclide-reaction-wise
    application_contributions, application_nested = determine_plot_type(application_contributions, plot_redundant_reactions)
    experiment_contributions, experiment_nested = determine_plot_type(experiment_contributions, plot_redundant_reactions)

    if application_nested != experiment_nested:
        raise ValueError("Application and experiment contributions must have the same nested structure")
    else:
        nested = application_nested # They are be the same, so arbitrarily choose one

    # Get the list of isotopes for which contributions are available
    isotopes = list(set(application_contributions.keys()).union(experiment_contributions.keys()))

    all_reactions = _add_missing_reactions_and_nuclides(application_contributions, experiment_contributions, isotopes, mode='contribution')

    # Now convert the contributions for the application and experiment into a list of x, y pairs for plotting
    contribution_pairs = []
    if nested:
        for isotope in isotopes:
            for reaction in all_reactions:
                contribution_pairs.append((application_contributions[isotope][reaction], \
                                           experiment_contributions[isotope][reaction]))
    else:
        for isotope in isotopes:
            contribution_pairs.append((application_contributions[isotope], experiment_contributions[isotope]))

    plotters = {
        'scatter': ScatterPlotter(integral_index_name, plot_redundant_reactions, nested, **kwargs),
        'interactive_scatter': InteractiveScatterPlotter(integral_index_name, plot_redundant_reactions, nested, **kwargs)
    }
    
    # Get the requested plotter
    plotter = plotters.get(plot_type)
    if plotter is None:
        raise ValueError("Unsupported plot type")

    # Create the plot and style it
    plotter.create_plot(contribution_pairs, isotopes, all_reactions)

    return plotter.fig


def perturbation_plot(points):
    """Plots the perturbation points for a given application-experiment pair for which the perturbation points have already
    been calculated
    
    Parameters
    ----------
    - points: list of tuple of ufloat, list of tuples containing the perturbation points for the application-experiment pair"""
    
    # Extracting data
    plotter = InteractivePerturbationScatterPlotter()
    plotter.create_plot(points)

    return plotter.get_plot()


def matrix_plot(plot_objects_array: np.ndarray, plot_type: str):
    """Creates a Dash app to display a matrix of plots from a numpy object array of figure objects."""
    if plot_type == 'interactive':
        return interactive_matrix_plot(plot_objects_array)
    elif plot_type == 'static':
        raise NotImplementedError("Static matrix plots are not yet supported")


import matplotlib.pyplot as plt
from .base_plotter import Plotter
import numpy as np

class BarPlotter(Plotter):
    def __init__(self, integral_index_name, plot_redundant=False, **kwargs):
        self.index_name = integral_index_name
        self.plot_redundant = plot_redundant

    def create_plot(self, contributions, nested):
        self.nested = nested
        self.fig, self.axs = plt.subplots()
        if nested:
            self.nested_barchart(contributions)
        else:
            self.barchart(contributions)

        self.style()

    def get_plot(self):
        return self.fig, self.axs
        
    def add_to_subplot(self, fig, position):
        return fig.add_subplot(position, sharex=self.axs, sharey=self.axs)
        
    def barchart(self, contributions):
        self.axs.bar(contributions.keys(), [contribution.n for contribution in contributions.values()],
            yerr=[contribution.s for contribution in contributions.values()], capsize=5, error_kw={'elinewidth': 0.5})

    def nested_barchart(self, contributions):
        # Colors for each reaction type
        num_reactions = len(next(iter(contributions.values())))
        cmap = plt.get_cmap('Set1')
        colors = cmap(np.linspace(0, 1, num_reactions))

        # Variables to hold the bar positions and labels
        indices = range(len(contributions))
        labels = list(contributions.keys())

        # Bottom offset for each stack
        bottoms_pos = [0] * len(contributions)
        bottoms_neg = [0] * len(contributions)

        color_index = 0
        for reaction in next(iter(contributions.values())).keys():
            values = [contributions[nuclide][reaction].n for nuclide in contributions]
            errs = [contributions[nuclide][reaction].s for nuclide in contributions]
            # Stacking positive values
            pos_values = [max(0, v) for v in values]
            neg_values = [min(0, v) for v in values]
            self.axs.bar(indices, pos_values, label=reaction, bottom=bottoms_pos, color=colors[color_index % len(colors)],
                    yerr=errs, capsize=5, error_kw={'capthick': 0.5})
            self.axs.bar(indices, neg_values, bottom=bottoms_neg, color=colors[color_index % len(colors)],
                    yerr=errs, capsize=5, error_kw={'capthick': 0.5})
            # Update the bottom positions
            bottoms_pos = [bottoms_pos[i] + pos_values[i] for i in range(len(bottoms_pos))]
            bottoms_neg = [bottoms_neg[i] + neg_values[i] for i in range(len(bottoms_neg))]
            color_index += 1

        # Adding 'effective' box with dashed border
        total_values = [sum(contributions[label][r].n for r in contributions[label]) for label in labels]
        for idx, val in zip(indices, total_values):
            self.axs.bar(idx, abs(val), bottom=0 if val > 0 else val, color='none', edgecolor='black', hatch='///', linewidth=0.5)

        self.axs.set_xticks(indices)
        self.axs.set_xticklabels(labels)
        self.axs.legend()

    def style(self):
        if self.plot_redundant and self.nested:
            title_text = f'Contributions to {self.index_name} (including redundant/irrelvant reactions)'
        else:
            title_text = f'Contributions to {self.index_name}'
        self.axs.set_ylabel(f"Contribution to {self.index_name}")
        self.axs.set_xlabel("Isotope")
        self.axs.grid(True, which='both', axis='y', color='gray', linestyle='-', linewidth=0.5)
        self.axs.set_title(title_text)


from uncertainties import ufloat, umath, unumpy
import numpy as np
from tsunami_ip_utils.readers import RegionIntegratedSdfReader, read_uncertainty_contributions
from tsunami_ip_utils._error import _unit_vector_uncertainty_propagation, _dot_product_uncertainty_propagation
from copy import deepcopy
from typing import List, Set, Tuple, Dict
from pathlib import Path
from uncertainties.core import Variable

def _calculate_E_from_sensitivity_vecs(application_vector: unumpy.uarray, experiment_vector: unumpy.uarray, 
                                      application_filename: Path=None, experiment_filename: Path=None, 
                                      uncertainties: str='automatic', experiment_norm: Variable=None, 
                                      application_norm: Variable=None) -> Variable:
    """Calculates the integral index: E given the sensitivity vectors for an application and an experiment. **NOTE**: the 
    application and experiment filenames are used to break the correlation between the application and experiment vectors 
    (that should not exist). This is because the vectors are only correlated if the application and experiment are the same.
    
    Parameters
    ----------
    application_vector
        Sensitivity vector for the application.
    experiment_vector
        Sensitivity vector for the experiment.
    application_filename
        Filename of the application sdf file (only needed for automatic uncertianty propagation).
    experiment_filename
        Filename of the experiment sdf file (only needed for automatic uncertianty propagation).
    uncertainties
        Type of error propagation to use. Default is 'automatic' which uses the uncertainties package.
        If set to 'manual', then manual error propagation is used which is generally faster
    experiment_norm
        Norm of the experiment vector. If not provided, it is calculated. This is mainly used for
        calculating E contributions, where the denominator is not actually the norm of the application and experiment
        vectors.
    application_norm
        Norm of the application vector. If not provided, it is calculated. This is mainly used for
        calculating E contributions, where the denominator is not actually the norm of the application and experiment
        vectors.
        
    Returns
    -------
        Similarity parameter between the application and the experiment"""
    
    norms_not_provided = ( experiment_norm == None ) and ( application_norm == None )
    if uncertainties == 'automatic': # Automatic error propagation with uncertainties package
        if application_filename == None or experiment_filename == None:
            raise ValueError("Application and experiment filenames must be provided for automatic error propagation")
        
        if norms_not_provided:
            application_norm = umath.sqrt(np.sum(application_vector**2))
            experiment_norm = umath.sqrt(np.sum(experiment_vector**2))
        
        application_unit_vector = application_vector / application_norm
        experiment_unit_vector = experiment_vector / experiment_norm

        # For some reason, the above introduces a correlation between the application and experiment vectors
        # which should only be the case if the application and the experiment are the same, so we manually 
        # break this correlation otherwise
        
        if application_filename != experiment_filename:
            # Break dependency to treat as independent
            application_unit_vector = np.array([ufloat(v.n, v.s) for v in application_unit_vector])
            experiment_unit_vector = np.array([ufloat(v.n, v.s) for v in experiment_unit_vector])

        dot_product = np.dot(application_unit_vector, experiment_unit_vector)

        E = dot_product
        return E

    else:
        # Manual error propagation

        # Same calculation for E
        if norms_not_provided:
            application_norm = umath.sqrt(np.sum(application_vector**2))
            experiment_norm = umath.sqrt(np.sum(experiment_vector**2))

        application_unit_vector = application_vector / application_norm
        experiment_unit_vector = experiment_vector / experiment_norm

        dot_product = np.dot(application_unit_vector, experiment_unit_vector)
        E = dot_product.n
        
        # ------------------------------------------
        # Now manually perform the error propagation
        # ------------------------------------------

        """Idea: propagate uncertainty of components of each sensitivity vector to the normalized sensitivty vector
        i.e. u / ||u|| = u^, v / ||v|| = v^
        
        E = u^ . v^

        Since u^ and v^ are completely uncorrelated, we may first calculate the uncertainty of u^ and v^ separately then
        use those to calculate the uncertainty of E."""

        # Calculate the uncertainties in the unit vectors
        application_unit_vector_error = _unit_vector_uncertainty_propagation(application_vector)
        experiment_unit_vector_error = _unit_vector_uncertainty_propagation(experiment_vector)

        # Construct application and experiment unit vectors as unumpy.uarray objects
        application_unit_vector = unumpy.uarray(unumpy.nominal_values(application_unit_vector), \
                                                    application_unit_vector_error)
        experiment_unit_vector = unumpy.uarray(unumpy.nominal_values(experiment_unit_vector), experiment_unit_vector_error)

        # Now calculate error in dot product to get the uncertainty in E
        E_uncertainty = _dot_product_uncertainty_propagation(application_unit_vector, experiment_unit_vector)

        return ufloat(E, E_uncertainty)


def _create_sensitivity_vector(sdfs: List[unumpy.uarray]) -> unumpy.uarray:
    """Creates a senstivity vector from all of the sensitivity profiles from a specific application or experiment
    by concatenating them together.

    Examples
    --------
    >>> sdfs = [ unumpy.uarray( [1, 2, 3], [0.1, 0.2, 0.3] ), unumpy.uarray( [4, 5, 6], [0.4, 0.5, 0.6] ) ]
    >>> create_sensitivity_vector(sdfs)
    array([1.0+/-0.1, 2.0+/-0.2, 3.0+/-0.3, 4.0+/-0.4, 5.0+/-0.5, 6.0+/-0.6],
          dtype=object)
    
    Parameters
    ----------
        List of sensitivity profiles for each nuclide-reaction pair in consideration.
    
    Returns
    -------
        Sensitivities from all of the sensitivity profiles combined into a single vector of length
        ``sum( [ len(sdf_profile) for sdf_profile in sdfs ] )``"""
    uncertainties = np.concatenate([unumpy.std_devs(sdf) for sdf in sdfs])
    senstivities = np.concatenate([unumpy.nominal_values(sdf) for sdf in sdfs])

    return unumpy.uarray(senstivities, uncertainties)


def calculate_E(application_filenames: List[str], experiment_filenames: List[str], reaction_type: str='all', 
                uncertainties: str='manual') -> np.ndarray:
    """Calculates the similarity parameter, E for each application with each available experiment given the application 
    and experiment sdf files
    
    Parameters
    ----------
    application_filenames
        Paths to the application sdf files.
    experiment_filenames
        Paths to the experiment sdf files.
    reaction_type 
        The type of reaction to consider in teh calculation of E. Default is ``'all``' which considers all 
        reactions.
    uncertainties 
        The type of uncertainty propagation to use. Default is ``'automatic'`` which uses the uncertainties
        package for error propagation. If set to ``'manual'``, then manual error propagation is used.
    
    Returns
    -------
        Similarity parameter for each application with each experiment, shape: ``(len(application_filenames), len(experiment_filenames))``"""

    # Read the application and experiment sdf files

    application_sdfs = [ RegionIntegratedSdfReader(filename).convert_to_dict() for filename in application_filenames ]
    experiment_sdfs  = [ RegionIntegratedSdfReader(filename).convert_to_dict() for filename in experiment_filenames ]

    # Create a matrix to store the similarity parameter E for each application with each experiment
    E_vals = unumpy.umatrix(np.zeros( ( len(experiment_sdfs), len(application_sdfs) ) ), \
                            np.zeros( ( len(experiment_sdfs), len(application_sdfs) ) ))
    
    # Now calculate the similarity parameter E for each application with each experiment
    for i, experiment in enumerate(experiment_sdfs):
        for j, application in enumerate(application_sdfs):
            # Now add missing data to the application and experiment dictionaries
            all_isotopes = set(application.sdf_data.keys()).union(set(experiment.sdf_data.keys()))
            _add_missing_reactions_and_nuclides(application.sdf_data, experiment.sdf_data, all_isotopes)
            
            # Sometimes the application and experiment dictionaries have different orders, which causes their sensitivity
            # profiles and hence sensitivity vectors to be different. To fix this, we need to sort the dictionaries

            # Sort the application by the experiment keys
            application.sdf_data = { isotope: { reaction: application.sdf_data[isotope][reaction] \
                                               for reaction in experiment.sdf_data[isotope] } \
                                                for isotope in experiment.sdf_data.keys() }

            application_profiles = application.get_sensitivity_profiles()
            experiment_profiles  = experiment.get_sensitivity_profiles()

            application_vector = _create_sensitivity_vector(application_profiles)
            experiment_vector  = _create_sensitivity_vector(experiment_profiles)

            E_vals[i, j] = _calculate_E_from_sensitivity_vecs(application_vector, experiment_vector, \
                                                             application_filenames[j], experiment_filenames[i], uncertainties)

    return E_vals

def _get_reaction_wise_E_contributions(application: dict, experiment: dict, isotope: str, all_reactions: List[str], 
                                      application_norm: Variable, experiment_norm: Variable) -> List[dict]:
    """Calculate contributions to the similarity parameter E for each reaction type for a given isotope.
    
    Parameters
    ----------
    application
        Dictionary of application sensitivity profiles.
    experiment
        Dictionary of experiment sensitivity profiles.
    isotope
        Isotope to consider.
    all_reactions
        List of all possible reaction types.
    application_norm
        Norm of the application sensitivity vector.
    experiment_norm
        Norm of the experiment sensitivity vector.
    
    Returns
    -------
        List of dictionaries containing the contribution to the similarity parameter E for each
        reaction type."""
    
    E_contributions = []
    for reaction in all_reactions:
        application_vector = application[isotope][reaction]['sensitivities']
        experiment_vector = experiment[isotope][reaction]['sensitivities']

        E_contribution = _calculate_E_from_sensitivity_vecs(application_vector, experiment_vector, uncertainties='manual', \
                                                           application_norm=application_norm, experiment_norm=experiment_norm)
        E_contributions.append({
            "isotope": isotope,
            "reaction_type": reaction,
            "contribution": E_contribution
        })

    return E_contributions

def _add_missing_reactions_and_nuclides(application: dict, experiment: dict, all_isotopes: List[str], 
                                       mode: str='sdfs') -> Set[str]:
    """Add missing reactions and nuclides to the application and experiment dictionaries with an sdf profile of all zeros.
    The set of all isotopes 
    **NOTE**: since dictionaries are passed by reference, this function does not return anything, but modifies the
    application and experiment dictionaries in place.
    
    Parameters
    ----------
    application
        Dictionary of application sensitivity profiles.
    experiment
        Dictionary of experiment sensitivity profiles.
    all_isotopes
        List of all isotopes in the application and experiment dictionaries.
    mode
        The mode to use for adding missing reactions and nuclides. Default is ``'sdfs'`` which adds sdf profiles to missing
        reactions and nuclides. If set to ``'contribution'``, then the contributions to the similarity parameter E are set to zero
        for missing nuclides and reactions.
    
    Returns
    -------
        Set of all isotopes in the application and experiment dictionaries."""
    # Whether or not the supplied data is only isotopes or includes reactions
    isotopes_only =  type( application[ list( application.keys() )[0] ] ) != dict

    if not isotopes_only:
        application_reactions = set([ key for isotope in application.keys() for key in application[isotope].keys() ])
        experiment_reactions = set([ key for isotope in experiment.keys() for key in experiment[isotope].keys() ])
        all_reactions = application_reactions.union(experiment_reactions)

        # Now, for any reactions that are in experiment but not application (and vice versa), they need to be added with an sdf
        # profile of all zeros

        # Get an arbitrary sdf profile to get the shape from
        first_application_nuclide = list(application.keys())[0]
        first_application_reaction = list(application[first_application_nuclide].keys())[0]
        if mode == 'sdfs':
            zero_data = {
                'sensitivities': application[first_application_nuclide][first_application_reaction]['sensitivities']
            }
        elif mode == 'contribution':
            zero_data = ufloat(0,0)

        # Now define a function used for updating the reactions for a given isotope
        def update_reactions(isotope):
            for reaction in all_reactions:
                if reaction not in isotope.keys():
                    isotope[reaction] = deepcopy(zero_data)
    else:
        # No reactions, only isotope totals
        all_reactions = []

        if mode == 'sdfs':
            zero_data = {
                'sensitivities': application[first_application_nuclide]['sensitivities']
            }
        elif mode == 'contribution':
            zero_data = ufloat(0,0)

    # ---------------------------------------
    # Now add missing reactions and nuclides
    # ---------------------------------------
    if not isotopes_only:
        for isotope in application.keys():
            # If reaction is missing for this isotope, add it with an sdf profile of all zeros
            update_reactions(application[isotope])

        for isotope in experiment.keys():
            # If reaction is missing for this isotope, add it with an sdf profile of all zeros
            update_reactions(experiment[isotope])

    # Now zero out nuclides that are not in the application or experiment
    for isotope in all_isotopes:
        if isotope not in application.keys():
            if not isotopes_only:
                application[isotope] = { reaction: deepcopy(zero_data) for reaction in all_reactions }
            else:
                application[isotope] = deepcopy(zero_data)
            
        if isotope not in experiment.keys():
            if not isotopes_only:
                experiment[isotope] = { reaction: deepcopy(zero_data) for reaction in all_reactions }
            else:
                experiment[isotope] = deepcopy(zero_data)

    return all_reactions

def _get_nuclide_and_reaction_wise_E_contributions(application: RegionIntegratedSdfReader, experiment: RegionIntegratedSdfReader
                                                   ) -> Tuple[List[dict], List[dict]]:
    """Calculate the contributions to the similarity parameter E for each nuclide and for each reaction type for a given
    application and experiment.
    
    Parameters
    ----------
    application
        Contains application sensitivity profile dictionaries.
    experiment
        Contains experiment sensitivity profile dictionaries.
    
    Returns
    -------
        * nuclide_wise_contributions
            List of dictionaries containing the contribution to the similarity parameter E.
            for each nuclide
        * nuclide_reaction_wise_contributions
            List of dictionaries containing the contribution to the similarity."""

    # First, extract the sensitivity vectors for the application and experiment
    
    # Calculate |S_A| and |S_E| to normalize the E contributions properly

    application_vector = _create_sensitivity_vector(application.get_sensitivity_profiles())
    experiment_vector = _create_sensitivity_vector(experiment.get_sensitivity_profiles())

    application_norm = umath.sqrt(np.sum(application_vector**2))
    experiment_norm = umath.sqrt(np.sum(experiment_vector**2))

    # Now convert the application and experiment sdf's to dictionaries keyed by nuclide and reaction type
    application = application.convert_to_dict().sdf_data
    experiment  = experiment.convert_to_dict().sdf_data

    nuclide_wise_contributions = []
    nuclide_reaction_wise_contributions = []

    # All isotopes in the application and experiment
    all_isotopes = set(application.keys()).union(set(experiment.keys()))

    # Since different nuclides can have different reactions, we need to consider all reactions for each nuclide (e.g. 
    # fissile isotopes will have fission reactions, while non-fissile isotopes will not)
    all_reactions = _add_missing_reactions_and_nuclides(application, experiment, all_isotopes)

    for isotope in all_isotopes:

        # For isotope-wise contribution, the sensitivity vector is all of the reaction sensitivities concatenated together
        application_vector = _create_sensitivity_vector([ application[isotope][reaction]['sensitivities'] \
                                                        for reaction in all_reactions ] )
        experiment_vector  = _create_sensitivity_vector([ experiment[isotope][reaction]['sensitivities'] \
                                                        for reaction in all_reactions ])

        E_isotope_contribution = _calculate_E_from_sensitivity_vecs(application_vector, experiment_vector, uncertainties='manual',
                                                                   application_norm=application_norm, \
                                                                    experiment_norm=experiment_norm)

        nuclide_wise_contributions.append({
            "isotope": isotope,
            "contribution": E_isotope_contribution
        })

        # For nuclide-reaction-wise contribution, we need to consider each reaction type
        nuclide_reaction_wise_contributions += \
            _get_reaction_wise_E_contributions(application, experiment, isotope, all_reactions, \
                                              application_norm, experiment_norm)
        
    return nuclide_wise_contributions, nuclide_reaction_wise_contributions


def calculate_E_contributions(application_filenames: List[str], experiment_filenames: List[str]
                              ) -> Tuple[unumpy.uarray, unumpy.uarray]:
    """Calculates the contributions to the similarity parameter E for each application with each available experiment 
    on a nuclide basis and on a nuclide-reaction basis.
    
    Parameters
    ----------
    application_filenames
        Paths to the application sdf files.
    experiment_filenames
        Paths to the experiment sdf files.
    
    Returns
    -------
        * E_contributions_nuclide
            Contributions to the similarity parameter E for each application with each
            experiment on a nuclide basis.
        * E_contributions_nuclide_reaction
            Contributions to the similarity parameter E for each application with
            each experiment on a nuclide-reaction basis."""
    
    application_sdfs = [ RegionIntegratedSdfReader(filename) for filename in application_filenames ]
    experiment_sdfs  = [ RegionIntegratedSdfReader(filename) for filename in experiment_filenames ]
    
    # Initialize np object arrays to store the E contributions
    E_nuclide_wise          = {'application': [], 'experiment': []}
    E_nuclide_reaction_wise = {'application': [], 'experiment': []}

    # Calculate contributions to E for each application
    for application in application_sdfs:
        nuclide_wise_contributions, nuclide_reaction_wise_contributions = \
            _get_nuclide_and_reaction_wise_E_contributions(application, application)

        E_nuclide_wise['application'].append(nuclide_wise_contributions)
        E_nuclide_reaction_wise['application'] = nuclide_reaction_wise_contributions

    # Calculate contributions to E for each experiment
    for experiment in experiment_sdfs:
        nuclide_wise_contributions, nuclide_reaction_wise_contributions = \
            _get_nuclide_and_reaction_wise_E_contributions(experiment, experiment)

        E_nuclide_wise['experiment'].append(nuclide_wise_contributions)
        E_nuclide_reaction_wise['experiment'] = nuclide_reaction_wise_contributions

    return E_nuclide_wise, E_nuclide_reaction_wise


def get_uncertainty_contributions(application_filenames: List[str], experiment_filenames: List[str]
                                   ) -> Tuple[ Dict[ str, List[unumpy.uarray] ], Dict[ str,  List[unumpy.uarray] ] ]:
    """Read the contributions to the uncertainty in :math:`k_{\\text{eff}}` (i.e. :math:`\\frac{dk}{k}`) for each 
    application with each available experiment on a nuclide basis and on a nuclide-reaction basis from the
    provided TSUNAMI-IP ``.out`` files.

    Parameters
    ----------
    application_filenames
        Paths to the application output (``.out``) files.
    experiment_filenames
        Paths to the experiment output (``.out``) files.

    Returns
    -------
        * uncertainty_contributions_nuclide
            List of contributions to the uncertainty in :math:`k_{\\text{eff}}` for each application and 
            each experiment on a nuclide basis. Keyed by ``'application'`` and ``'experiment'``.
        * uncertainty_contributions_nuclide_reaction
            List of contributions to the uncertainty in :math:`k_{\\text{eff}}` for each application and each experiment 
            on a nuclide-reaction basis. Keyed by ``'application'`` and ``'experiment'``."""
    
    dk_over_k_nuclide_wise = {
        'application': np.empty( len(application_filenames), dtype=object ),
        'experiment': np.empty( len(experiment_filenames), dtype=object )    
    }
    dk_over_k_nuclide_reaction_wise = {
        'application': np.empty( len(application_filenames), dtype=object ),
        'experiment': np.empty( len(experiment_filenames), dtype=object )    
    }

    for i, application_filename in enumerate(application_filenames):
        dk_over_k_nuclide_wise['application'][i], dk_over_k_nuclide_reaction_wise['application'][i] = \
            read_uncertainty_contributions(application_filename)

    for i, experiment_filename in enumerate(experiment_filenames):
        dk_over_k_nuclide_wise['experiment'][i], dk_over_k_nuclide_reaction_wise['experiment'][i] = \
            read_uncertainty_contributions(experiment_filename)
            
    return dk_over_k_nuclide_wise, dk_over_k_nuclide_reaction_wise